# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


from .classes import Expression, \
    Program, Node, Identifier, Factorize, Hyperedge, Variable, Sizing, State, \
    Action, Auxiliary, MDPObjective, MDP, Time, Parameter
from .classes.error import RedefinitionError

import numpy as np  # type: ignore
from .utils import error_, get_branch_in_nested_dict, \
    update_branch_in_nested_dict, get_layer_in_nested_dict
import time as t


def semantic(program: Program) -> tuple:
    # CHECKS to do -> Check if there are no constraints
    time = program.get_time()
    check_time_horizon(time)
    time_value = time.get_value()

    node_list = program.get_nodes()
    link_list = program.get_links()
    check_layer_repetition(node_list, link_list)

    definitions = dict()
    time_parameter = Parameter("T", Expression("literal", time_value))
    time_parameter.set_value([time_value])

    definitions["T"] = time_parameter
    global_param = program.get_dict_global_parameters()
    parameter_evaluation(global_param, definitions)
    definitions["global"] = global_param
    program_variables_dict = {}

    global_index = 0

    if not node_list:
        error_("ERROR: No node defined")

    check_objective_exists(node_list)

    global_index = \
        check_and_extend_parameters_variables_in_nodes(node_list,
                                                       program_variables_dict,
                                                       definitions, [],
                                                       global_index)
    check_and_extend_parameters_in_edges(link_list, definitions, [])

    for node in node_list:
        node_name = node.get_name()
        recursive_node_name_addition(node)
        current_definition = \
            {"T": definitions["T"], "global": definitions["global"],
             node_name: definitions[node_name]}
        recursive_check_definition_node(node, current_definition,
                                        program_variables_dict, definitions)

    for link in link_list:
        edge_name = link.get_name()
        add_link_names(link)
        current_definition = \
            {"T": definitions["T"], "global": definitions["global"],
             edge_name: definitions[edge_name]}
        check_definition_link(link, program_variables_dict, current_definition)

    program.set_nb_var_index(global_index)
    program.set_variables_dict(program_variables_dict)
    program.set_global_parameters(global_param)

    return program, program_variables_dict, definitions


def check_constraint_exists(node_list: list, hyperlink_list: list):
    constraint_exist = False
    for element in node_list + hyperlink_list:
        constraints = element.get_constraints()
        if constraints:
            constraint_exist = True
            break
    if not constraint_exist:
        error_("ERROR: There is no constraint defined")


def check_objective_exists(node_list: list):
    objective_defined = objective_exists(node_list)
    if not objective_defined:
        error_("ERROR: There is no objective defined")


def objective_exists(node_list: list) -> bool:
    objective_defined = False
    for node in node_list:
        if node.get_objectives():
            objective_defined = True
            break

        objective_defined = objective_exists(node.get_sub_nodes())
        if objective_defined:
            break
    return objective_defined


def apply_changes_variables(dictionary_of_variables, variables_changes):
    for variable_id, variable_type, variable_line in variables_changes:
        if variable_id not in dictionary_of_variables:
            error_("ERROR : variable id : " + str(variable_id)
                   + " does not exist at line: " + str(variable_line))
        else:
            variable_considered = dictionary_of_variables[variable_id]
            variable_considered.reset_type(variable_type)


def apply_changes_parameters(dictionary_of_parameters, parameters_changes):
    for new_parameter in parameters_changes:
        parameter_name = new_parameter.get_name()
        if parameter_name not in dictionary_of_parameters:
            error_("ERROR : parameter id " + str(parameter_name)
                   + " does not exist at line: "
                   + str(new_parameter.get_line()))
        else:
            previous_parameter = dictionary_of_parameters[parameter_name]
            new_parameter_type = new_parameter.get_type()
            previous_parameter_type = previous_parameter.get_type()
            if previous_parameter_type != new_parameter_type:
                error_("ERROR : the redefinition of " + str(parameter_name)
                       + " does not match its previous type at lien: "
                       + str(new_parameter.get_line()))

            if new_parameter.get_number_of_values() != \
                    previous_parameter.get_number_of_values():
                print("WARNING : unmatching length in redefinition"
                      " of parameter at line: "
                      + str(new_parameter.get_line()))
            dictionary_of_parameters[parameter_name] = new_parameter


def check_and_extend_parameters_variables_in_nodes(nodes_list: list,
                                                   program_variables_dict: dict,
                                                   definitions: dict,
                                                   depth_list: list,
                                                   global_index: int):
    for node in nodes_list:
        node_name = node.get_name()
        variables_changes = node.get_variables_changes()
        parameters_changes = node.get_parameters_changes()
        # Retrieve node parameter dictionary
        node_parameters = node.get_dictionary_parameters()
        node_variables = node.get_dictionary_variables(get_id=False)
        node_expressions = node.get_dictionary_expressions()
        apply_changes_variables(node_variables, variables_changes)
        apply_changes_parameters(node_parameters, parameters_changes)
        match_dictionaries(node_parameters, node_variables, node_expressions)
        _, flat_branch_dictionary = \
            get_branch_in_nested_dict(definitions, depth_list, not_lower=True)
        flat_branch_dictionary["global"] = definitions["global"]
        flat_branch_dictionary["T"] = definitions["T"]
        parameter_evaluation(node_parameters, flat_branch_dictionary)

        flat_branch_dictionary[node_name] = node_parameters
        node.set_parameter_dict(node_parameters)
        _, definitions = \
            update_branch_in_nested_dict(definitions, depth_list.copy(),
                                         node_name, node_parameters)
        _, program_variables_dict = \
            update_branch_in_nested_dict(program_variables_dict,
                                         depth_list.copy(),
                                         node_name,
                                         node_variables.copy())
        depth_list.append(node_name)
        global_index = \
            check_and_extend_parameters_variables_in_nodes(
                node.get_sub_nodes(),
                program_variables_dict,
                definitions, depth_list, global_index)
        check_and_extend_parameters_in_edges(node.get_sub_hyperedges(),
                                             definitions, depth_list)
        _, nested_nodes_variables = get_layer_in_nested_dict(
            program_variables_dict, depth_list, only_dict=True)
        depth_list.remove(node_name)
        global_index = set_size_variables(node_variables,
                                          flat_branch_dictionary,
                                          global_index,
                                          nested_nodes_variables)
        check_names_repetitions(node.get_constraints())
        check_names_repetitions(node.get_objectives())

    return global_index


def check_and_extend_parameters_in_edges(edge_list: list,
                                         definitions: dict,
                                         depth_list: list):
    for edge in edge_list:
        edge_name = edge.get_name()
        parameters_changes = edge.get_parameters_changes()
        names_changes = edge.get_names_changes()
        # Retrieve node parameter dictionary
        edge_parameters = edge.get_dictionary_parameters()
        apply_changes_parameters(edge_parameters, parameters_changes)
        change_node_names_in_edge(edge, names_changes)
        _, flat_branch_dictionary = get_branch_in_nested_dict(definitions,
                                                              depth_list,
                                                              not_lower=True)
        flat_branch_dictionary["global"] = definitions["global"]
        flat_branch_dictionary["T"] = definitions["T"]
        parameter_evaluation(edge_parameters, flat_branch_dictionary)
        edge.set_parameter_dict(edge_parameters)
        _, definitions = update_branch_in_nested_dict(definitions,
                                                      depth_list.copy(),
                                                      edge_name,
                                                      edge_parameters)


def change_node_names_in_edge(edge, names_changes):
    changes_dictionary = {}
    for lhs_id, rhs_id, _ in names_changes:
        changes_dictionary[lhs_id] = rhs_id

    if changes_dictionary != {}:
        constraints = edge.get_constraints()
        for constraint in constraints:
            leaves = constraint.get_leafs()
            for leaf in leaves:
                seed = leaf.get_name()
                if type(seed) == Identifier:
                    current_node_name = seed.get_node_name()
                    if current_node_name in changes_dictionary:
                        seed.set_node_name(
                            changes_dictionary[current_node_name])


def check_program_linearity(program, variables, definitions):
    definition_accumulator_dict = \
        {"T": definitions["T"], "global": definitions["global"]}
    nodes = program.get_nodes()
    hyperedges = program.get_links()
    recursive_application_on_nodes_hyperedges(nodes, hyperedges, variables,
                                              definitions,
                                              definition_accumulator_dict,
                                              check_expressions_dependency,
                                              check_expressions_dependency_link)


def recursive_application_on_nodes_hyperedges(nodes,
                                              hyperedges,
                                              variables: dict,
                                              definitions,
                                              accumulator_dict,
                                              function_nodes,
                                              function_hyperedge):
    for node in nodes:
        node_name = node.get_name()
        sub_nodes = node.get_sub_nodes()
        sub_edges = node.get_sub_hyperedges()
        accumulator_dict[node_name] = definitions[node_name]
        function_nodes(node, variables, accumulator_dict)
        recursive_application_on_nodes_hyperedges(sub_nodes, sub_edges,
                                                  variables[node_name],
                                                  definitions[node_name],
                                                  accumulator_dict,
                                                  function_nodes,
                                                  function_hyperedge)
        accumulator_dict.pop(node_name)

    for link in hyperedges:
        function_hyperedge(link, variables, definitions)


def factorize_program(program, variables, definitions):
    definition_accumulator_dict = \
        {"T": definitions["T"], "global": definitions["global"]}
    nodes = program.get_nodes()
    hyperedges = program.get_links()
    recursive_application_on_nodes_hyperedges(nodes, hyperedges,
                                              variables, definitions,
                                              definition_accumulator_dict,
                                              factorize_node,
                                              factorize_hyperedge)


def factorize_node(node, variables, definitions):
    name = node.get_name()
    start_time = t.time()
    constraints = node.get_constraints()
    objectives = node.get_objectives()
    constraint_factors = []
    for constraint in constraints:
        factor = Factorize(constraint)
        factor.factorize_constraint(variables, definitions)
        constraint_factors.append(factor)

    node.set_constraint_factors(constraint_factors)

    objective_factors = []
    for objective in objectives:
        factor = Factorize(objective)
        factor.factorize_objective(variables, definitions)
        objective_factors.append(factor)
    node.set_objective_factors(objective_factors)

    print("Check variables of node %s : --- %s seconds ---"
          % (name, t.time() - start_time))


def factorize_hyperedge(edge, variables, definitions):
    name = edge.get_name()
    constraints = edge.get_constraints()
    constraint_factors = []

    for constraint in constraints:
        factor = Factorize(constraint)
        factor.factorize_constraint(variables, definitions)
        constraint_factors.append(factor)
    edge.set_constraint_factors(constraint_factors)
    start_time = t.time()
    print("Check hyperlink %s : --- %s seconds ---"
          % (name, t.time() - start_time))


def check_expressions_dependency_link(link: Hyperedge,
                                      variables: dict,
                                      parameters_obj: dict):
    constraints = link.get_constraints()
    for cons in constraints:

        rhs = cons.get_rhs()
        lhs = cons.get_lhs()

        var_in_right = predicate_variables_in_expression(rhs, variables)
        var_in_left = predicate_variables_in_expression(lhs, variables)

        if var_in_right is False and var_in_left is False:
            error_('No variable in constraint at line ' + str(cons.get_line()))

        check_linear(rhs, variables, parameters_obj)
        check_linear(lhs, variables, parameters_obj)


def check_expressions_dependency(node: Node,
                                 variables: dict,
                                 parameters_obj: dict):
    """
    check_expressions_dependancy function : checks the expressions inside a node
    INPUT:  node -> Node object
            variables -> dictionary of <name,identifier> objects
            parameters -> dictionary of <name,array> objects
    OUTPUT: None
    """
    constraints = node.get_constraints()
    for cons in constraints:

        rhs = cons.get_rhs()
        lhs = cons.get_lhs()

        var_in_right = predicate_variables_in_expression(rhs, variables)
        var_in_left = predicate_variables_in_expression(lhs, variables)

        if var_in_right is False and var_in_left is False:
            error_('No variable in constraint at line ' + str(cons.get_line()))

        check_linear(rhs, variables, parameters_obj)
        check_linear(lhs, variables, parameters_obj)

    objectives = node.get_objectives()
    for obj in objectives:

        expr = obj.get_expression()

        contains_var = predicate_variables_in_expression(expr, variables)

        if contains_var is False:
            error_(
                'Objective only depends on constants not on variable at line '
                + str(expr.get_line()))
        check_linear(expr, variables, parameters_obj)


def check_linear(expression: Expression,
                 variables: dict,
                 parameters: dict) -> bool:
    """
    check_linear function : checks if an expression is linear with respect
                            with respect to the variables
    INPUT:  expression -> expression object to check
            variables -> dictionary of <name,identifier> objects
            parameters -> dictionary of <name,array> objects
    OUTPUT: bool -> boolean value if it depends on a variable
    """

    e_type = expression.get_type()
    nb_child = expression.get_nb_children()
    children = expression.get_children()

    if e_type == 'literal':

        if nb_child != 0:
            error_(
                "INTERNAL ERROR : literal expression must "
                "have zero child, got "
                + str(nb_child) +
                " check internal parser")
    elif e_type == 'u-':
        if nb_child != 1:
            error_("INTERNAL ERROR : unary minus operator "
                   "must have one child, got "
                   + str(nb_child) +
                   " check internal parser")
        lin1 = check_linear(children[0], variables, parameters)
        if lin1 is False:
            error_("Non linearity in expression : "
                   + str(children[0])
                   + " only linear problems are accepted at line "
                   + str(children[0].get_line()))
    elif e_type == 'sum':

        if nb_child != 1:
            error_("INTERNAL ERROR : sum operator must have one child, got "
                   + str(nb_child) + " check internal parser")
        time_int = expression.get_time_interval()
        time_var = time_int.get_index_name()
        if time_var in parameters:
            error_("Redefinition of " + str(time_int) + " at line : "
                   + str(expression.get_line()))
        parameters[time_var] = None

        lin1 = check_linear(children[0], variables, parameters)
        parameters.pop(time_var)
        if lin1 is False:
            error_("Non linearity in expression : " + str(children[0])
                   + " only linear problems are accepted at line "
                   + str(children[0].get_line()))
    else:

        if nb_child != 2:
            error_("INTERNAL ERROR : binary operators "
                   "must have two children, got "
                   + str(nb_child) +
                   " check internal parser")
        term1 = predicate_variables_in_expression(children[0], variables)
        term2 = predicate_variables_in_expression(children[1], variables)
        if e_type == "-" or e_type == '+':

            if term1 is True:

                lin1 = check_linear(children[0], variables, parameters)
                if lin1 is False:
                    error_("Non linearity in expression : " + str(children[0]) +
                           " only linear problems are accepted at line "
                           + str(children[0].get_line()))
            if term2 is True:

                lin2 = check_linear(children[1], variables, parameters)
                if lin2 is False:
                    error_("Non linearity in expression : " + str(children[1]) +
                           " only linear problems are accepted at line "
                           + str(children[0].get_line()))
        elif e_type == "*" or e_type == "/":

            if term2 is True and term1 is True:
                string = "Operation '" + str(e_type) + \
                         "' between two expressions containing " \
                         "variables leading to a non linearity at line " \
                         + str(children[0].get_line()) + "\n"
                string += "Namely Expression 1 : " + str(children[0]) \
                          + " and Expression 2 : " + str(children[1])
                error_(string)
            if term2 is True and e_type == "/":
                string = "A variable in the denominator of " \
                         "a division leads to a Non linearity at line " \
                         + str(children[0].get_line())
                error_(string)
            if term1 is True:

                lin1 = check_linear(children[0], variables, parameters)
                if lin1 is False:
                    error_("Non linearity in expression : " + str(children[0]) +
                           " only linear problems are accepted at line "
                           + str(children[0].get_line()))
        elif e_type == "**":

            if term1 is True or term2 is True:
                string = "Operation '" + str(e_type) + \
                         "' between one expression containing variables " \
                         "leading to a non linearity at line " \
                         + str(children[0].get_line()) + "\n"
                string += "Namely Expression 1 : " + str(children[0]) \
                          + " and Expression 2 : " + str(children[1])
                error_(string)
        elif e_type == "mod":

            string = "Non linearity, modulo operator is " \
                     "not allowed on variables at line " \
                     + str(children[0].get_line()) + "\n"
            error_(string)
        else:

            error_("INTERNAL ERROR : unknown type '"
                   + str(e_type) + "' check internal parser")
    return True


def check_mdp(program, variables: dict, definitions: dict):
    nodes = program.get_nodes()
    hyperlinks = program.get_links()
    check_variables_type_sizes(variables, definitions)

    for node in nodes:
        name = node.get_name()
        start_time = t.time()
        replace_parameters_node(node, definitions)
        check_expressions_dependancy_rl(node, variables_dict=variables)
        print("Check variables of node %s : --- %s seconds ---"
              % (name, t.time() - start_time))

    for link in hyperlinks:
        name = link.get_name()
        start_time = t.time()
        replace_parameters_hyperlinks(link, definitions)
        check_link_rl(link, variables)
        print("Check hyperlink %s : --- %s seconds ---"
              % (name, t.time() - start_time))

    check_all_variables(variables)


def replace_parameters_hyperlinks(hyperlink: Hyperedge, definitions):
    constraints = hyperlink.get_constraints()
    for constraint in constraints:
        rhs: Expression = constraint.get_rhs()
        lhs: Expression = constraint.get_lhs()
        rhs.replace_basic_parameters(definitions)
        lhs.replace_basic_parameters(definitions)


def replace_parameters_node(node: Node, definitions):
    constraints = node.get_constraints()
    objectives = node.get_objectives()
    for constraint in constraints:
        rhs: Expression = constraint.get_rhs()
        lhs: Expression = constraint.get_lhs()
        rhs.replace_basic_parameters(definitions)
        lhs.replace_basic_parameters(definitions)

    for objective in objectives:
        obj_expression: Expression = objective.get_expression()
        obj_expression.replace_basic_parameters(definitions)


def check_definition_link(link: Hyperedge, variables_dict: dict,
                          parameters_dict: dict, parent_hood=None):
    if parent_hood is None:
        parent_hood = []

    constraints_list = link.get_constraints()
    link_name = link.get_name()

    for constraint in constraints_list:
        index_var = constraint.get_index_var()
        if link_name in parameters_dict and \
                index_var in parameters_dict[link_name]:
            error_("ERROR: index name " + str(index_var) + " at line "
                   + str(constraint.get_line())
                   + " is already used to define parameter "
                   + str(parameters_dict[link_name][index_var]))

        reserved = [index_var]
        lhs = constraint.get_lhs()
        check_definition_expression(lhs, variables_dict, parameters_dict,
                                    reserved, in_node_name_parameter=link_name,
                                    variable_type="external",
                                    parent_hood=parent_hood)
        rhs = constraint.get_rhs()
        check_definition_expression(rhs, variables_dict, parameters_dict,
                                    reserved, in_node_name_parameter=link_name,
                                    variable_type="external",
                                    parent_hood=parent_hood)


def recursive_check_definition_node(node: Node, current_definitions,
                                    variables_dict: dict, parameters_dict: dict,
                                    parent_hood=None):
    if parent_hood is None:
        parent_hood = []
    node_name = node.get_name()
    sub_nodes, sub_edges = node.get_sub_nodes(), node.get_sub_hyperedges()
    check_definition(node, variables_dict, current_definitions,
                     parent_hood=parent_hood)
    variables_dict = variables_dict[node_name]
    parameters_dict = parameters_dict[node_name]
    parent_hood.append(node_name)
    for sub_node in sub_nodes:
        sub_node_name = sub_node.get_name()
        current_definitions[sub_node_name] = parameters_dict[sub_node_name]
        sub_node_variables = {sub_node_name: variables_dict[sub_node_name]}
        recursive_check_definition_node(sub_node, current_definitions,
                                        sub_node_variables,
                                        parameters_dict, parent_hood)
        current_definitions.pop(sub_node_name)

    for edge in sub_edges:
        edge_name = edge.get_name()
        current_definitions[edge_name] = parameters_dict[edge_name]
        check_definition_link(edge, variables_dict,
                              current_definitions, parent_hood)
        current_definitions.pop(edge_name)
    parent_hood.pop(-1)


def check_definition(node: Node, variables_dict: dict,
                     parameters_dict: dict, parent_hood):
    objectives_list = node.get_objectives()
    constraints_list = node.get_constraints()
    node_name = node.get_name()

    for constraint in constraints_list:
        index_var = constraint.get_index_var()

        if node_name in parameters_dict and \
                index_var in parameters_dict[node_name]:
            error_("ERROR: index name " + str(index_var)
                   + " at line " + str(constraint.get_line())
                   + " is already used to define parameter "
                   + str(parameters_dict[node_name][index_var]))

        elif node_name in variables_dict and \
                index_var in variables_dict[node_name]:
            error_("ERROR: index name " + str(index_var) + " at line "
                   + str(constraint.get_line())
                   + " is already used to define parameter "
                   + str(variables_dict[node_name][index_var]))

        reserved = [index_var]
        lhs = constraint.get_lhs()
        check_definition_expression(lhs, variables_dict,
                                    parameters_dict, reserved,
                                    in_node_name_variable=node_name,
                                    in_node_name_parameter=node_name,
                                    parent_hood=parent_hood)
        rhs = constraint.get_rhs()
        check_definition_expression(rhs, variables_dict, parameters_dict,
                                    reserved,
                                    in_node_name_variable=node_name,
                                    in_node_name_parameter=node_name,
                                    parent_hood=parent_hood)

    for objective in objectives_list:
        expr = objective.get_expression()
        index_var = objective.get_index_var()

        if node_name in parameters_dict and \
                index_var in parameters_dict[node_name]:
            error_("ERROR: index name " + str(index_var) + " at line "
                   + str(objective.get_line())
                   + " is already used to define parameter "
                   + str(parameters_dict[node_name][index_var]))

        elif node_name in variables_dict and \
                index_var in variables_dict[node_name]:
            error_("ERROR: index name " + str(index_var) + " at line "
                   + str(objective.get_line())
                   + " is already used to define parameter "
                   + str(variables_dict[node_name][index_var]))

        reserved = [index_var]
        check_definition_expression(expr, variables_dict,
                                    parameters_dict, reserved,
                                    in_node_name_variable=node_name,
                                    in_node_name_parameter=node_name,
                                    parent_hood=parent_hood)


def check_definition_expression(expr: Expression, variables_dict: dict,
                                parameters_dict: dict, reserved=None,
                                variables_allowed=True,
                                in_node_name_variable="",
                                in_node_name_parameter="",
                                variable_type="", parent_hood=None):
    if parent_hood is None:
        parent_hood = []
    if reserved is None:
        reserved = []
    leaves = expr.get_leafs()

    for leaf in leaves:
        in_node_name = ""
        is_reserved = False
        seed = leaf.get_name()

        if type(seed) == Identifier:
            defined = False
            seed_identifier: Identifier = seed
            seed_node_name = seed_identifier.get_node_name()
            seed_id_name = seed_identifier.get_name()

            if seed_node_name == "" and \
                    (seed_id_name in reserved or seed_id_name == "T"):
                defined = True
                is_reserved = True

            elif seed_node_name in variables_dict and \
                    seed_id_name in variables_dict[seed_node_name]:
                defined = True
                in_node_name = in_node_name_variable
                variable_referenced: Variable = \
                    variables_dict[seed_node_name][seed_id_name]
                variable_referenced_type = variable_referenced.get_type()
                identifier_referenced = variable_referenced.get_identifier()
                identifier_referenced_type = identifier_referenced.get_type()

                if not variables_allowed:
                    error_("ERROR: variables are not "
                           "allowed in brackets at line "
                           + str(leaf.get_line()))

                if seed_identifier.get_type() != identifier_referenced_type:
                    error_("ERROR: Unmatching type between definition of "
                           + str(identifier_referenced) +
                           " and usage " + str(seed_identifier) + " at line "
                           + str(seed.get_line()))

                if variable_type != "" and \
                        variable_type != variable_referenced_type:
                    error_("ERROR: Only " + str(variable_type)
                           + " are accepted at line : " + str(seed.get_line())
                           + " got " + str(variable_referenced) + " of type "
                           + str(variable_referenced_type))

            elif seed_node_name in parameters_dict and \
                    seed_id_name in parameters_dict[seed_node_name]:
                parameter_referenced: Parameter = \
                    parameters_dict[seed_node_name][seed_id_name]
                parameter_referenced_type = parameter_referenced.get_type()
                defined = True
                in_node_name = in_node_name_parameter
                identifier_referenced_type = ""
                if parameter_referenced_type == "expression":
                    identifier_referenced_type = "basic"
                elif parameter_referenced_type == "table":
                    identifier_referenced_type = "assign"

                if seed_identifier.get_type() != identifier_referenced_type:
                    error_("ERROR: Unmatching type between definition of "
                           + str(seed_id_name) +
                           " and usage " + str(seed_identifier) + " at line "
                           + str(seed.get_line()))

            if not defined:
                error_("ERROR: Unknown identifier "
                       + str(seed_identifier)
                       + " at line " + str(leaf.get_line()))

            elif not is_reserved and \
                    in_node_name != "" and \
                    seed_node_name != "global" and \
                    in_node_name != seed_node_name and \
                    seed_node_name not in parent_hood:
                print(in_node_name)
                print(parameters_dict)
                print(variables_dict)
                error_("ERROR: referencing variables "
                       "or parameters defined in node "
                       + str(seed_node_name) +
                       " inside node " + str(in_node_name)
                       + " is not allowed at line " + str(expr.get_line()))

            expression_in_brackets = seed_identifier.get_expression()
            if expression_in_brackets is not None:
                check_definition_expression(expression_in_brackets,
                                            variables_dict,
                                            parameters_dict,
                                            reserved,
                                            variables_allowed=False,
                                            in_node_name_variable=in_node_name_variable,
                                            in_node_name_parameter=in_node_name_parameter,
                                            variable_type=variable_type)

        if leaf.get_type() == "sum":
            sum_expr: Expression = leaf
            sum_children = sum_expr.get_children()
            time_interval = sum_expr.get_time_interval()
            index_name = time_interval.get_index_name()

            if in_node_name in variables_dict and \
                    index_name in variables_dict[in_node_name]:
                error_("ERROR: Redefinition of variable name by index "
                       + str(index_name) +
                       " at line " + str(leaf.get_line()))

            if in_node_name in parameters_dict and \
                    index_name in parameters_dict[in_node_name]:
                error_("ERROR: Redefinition of parameter name by index "
                       + str(index_name) +
                       " at line " + str(leaf.get_line()))

            reserved_sum = reserved + [index_name]
            for child in sum_children:
                check_definition_expression(child, variables_dict,
                                            parameters_dict, reserved_sum,
                                            variables_allowed,
                                            in_node_name_variable,
                                            in_node_name_parameter,
                                            variable_type)


def add_link_names(link: Hyperedge):
    constraints_list = link.get_constraints()

    for constraint in constraints_list:
        lhs = constraint.get_lhs()
        index_var = constraint.get_index_var()

        reserved = [index_var]
        add_node_names_expression(lhs, link, reserved)
        rhs = constraint.get_rhs()
        add_node_names_expression(rhs, link, reserved)


def recursive_node_name_addition(node):
    add_node_names(node)
    for sub_node in node.get_sub_nodes():
        recursive_node_name_addition(sub_node)

    for sub_edge in node.get_sub_hyperedges():
        add_link_names(sub_edge)


def add_node_names(node: Node):
    objectives_list = node.get_objectives()
    constraints_list = node.get_constraints()
    expressions_list = node.get_expressions()

    for expr_name, expression, line_number in expressions_list:
        add_node_names_expression(expression, node, [])

    for constraint in constraints_list:
        condition = constraint.get_condition()
        time_interval = constraint.get_time_interval()
        if time_interval is not None:
            begin = time_interval.get_begin()
            step = time_interval.get_step()
            end = time_interval.get_end()
            add_node_names_expression(begin, node, [])
            add_node_names_expression(step, node, [])
            add_node_names_expression(end, node, [])

        index_var = constraint.get_index_var()
        reserved = [index_var]

        if condition is not None:
            condition_lhs, condition_rhs = condition.get_children()
            add_node_names_expression(condition_lhs, node, reserved)
            add_node_names_expression(condition_rhs, node, reserved)

        lhs = constraint.get_lhs()
        add_node_names_expression(lhs, node, reserved)
        rhs = constraint.get_rhs()
        add_node_names_expression(rhs, node, reserved)

    for objective in objectives_list:
        expr = objective.get_expression()
        condition = objective.get_condition()
        time_interval = objective.get_time_interval()
        if time_interval is not None:
            begin = time_interval.get_begin()
            step = time_interval.get_step()
            end = time_interval.get_end()
            add_node_names_expression(begin, node, [])
            add_node_names_expression(step, node, [])
            add_node_names_expression(end, node, [])
        index_var = objective.get_index_var()
        reserved = [index_var]
        if condition is not None:
            condition_lhs, condition_rhs = condition.get_children()
            add_node_names_expression(condition_lhs, node, reserved)
            add_node_names_expression(condition_rhs, node, reserved)

        add_node_names_expression(expr, node, reserved)


def add_node_names_expression(expr: Expression, node,
                              reserved_identifiers=None):
    if reserved_identifiers is None:
        reserved_identifiers = []
    leaves = expr.get_leafs()
    reserved = ["T"] + reserved_identifiers
    for leaf in leaves:
        seed = leaf.get_name()
        if type(seed) == Identifier:
            identifier: Identifier = seed
            if identifier.get_node_name() == "" and \
                    identifier.get_name() not in reserved:
                identifier.set_node(node)
            if identifier.get_type() == "assign":
                add_node_names_expression(identifier.get_expression(), node,
                                          reserved_identifiers)
        if leaf.get_type() == "sum":
            sum_expr: Expression = leaf
            sum_children = sum_expr.get_children()
            time_interval = sum_expr.get_time_interval()
            index_name = time_interval.get_index_name()
            reserved_sum = reserved + [index_name]
            for child in sum_children:
                add_node_names_expression(child, node, reserved_sum)


def check_time_horizon(time_horizon: Time):
    variable_name = time_horizon.get_name()
    line = time_horizon.get_line()
    if variable_name != "T":
        error_("Semantic error:" + str(
            line) + ": Use \"T\"" + " as a symbol for the time horizon. \""
               + str(variable_name) + "\" is not allowed")

    time_value = time_horizon.get_value()
    if type(time_value) == float and time_value.is_integer() is False:

        time_value = int(round(time_value))
        print("WARNING: the time horizon considered is not an int")
        print("The time horizon was rounded to " + str(time_value))
    elif type(time_value) == float:

        time_value = int(time_value)
    if time_value < 0:

        error_("ERROR: the chosen time horizon is negative.")
    elif time_value == 0:

        print("WARNING: the time horizon considered is 0")

    time_horizon.set_value(time_value)


def convert_expression_identifiers_to_basic(expression: Expression):
    leaves = expression.get_leafs()
    for leaf in leaves:
        seed = leaf.get_name()
        seed_type = type(seed)
        if seed_type == Identifier:
            identifier: Identifier = seed
            identifier_type = identifier.get_type()
            if identifier_type == "assign":
                identifier.set_expression(None)
                identifier.set_type("basic")


def convert_to_mdp(program, program_variables_dict):
    list_states = []
    list_auxiliaries = []
    list_actions = []
    list_sizing = []
    list_mdpobjectives = []
    nodes = program.get_nodes()

    for node_name in program_variables_dict.keys():
        node_variables_dict = program_variables_dict[node_name]
        for variable_name in node_variables_dict.keys():
            variable = node_variables_dict[variable_name]
            var_option = variable.get_option()
            if var_option == "state":
                variable_dynamics = variable.get_dynamic()
                variable_initial_constraint = variable.get_initial_constraint()

                if variable_dynamics is not None:
                    convert_expression_identifiers_to_basic(variable_dynamics)

                if variable_initial_constraint is not None:
                    convert_expression_identifiers_to_basic(
                        variable_initial_constraint)

                state_var = State(variable_name, node_name, variable_dynamics,
                                  variable_initial_constraint)
                list_states.append(state_var)

            elif var_option == "sizing":
                variable_lower_constraint = variable.get_lower_constraint()
                variable_upper_constraint = variable.get_upper_constraint()
                if variable_lower_constraint is not None:
                    convert_expression_identifiers_to_basic(
                        variable_lower_constraint)
                if variable_upper_constraint is not None:
                    convert_expression_identifiers_to_basic(
                        variable_upper_constraint)
                sizing_var = Sizing(variable_name, node_name,
                                    variable_lower_constraint,
                                    variable_upper_constraint)
                list_sizing.append(sizing_var)

            elif var_option == "action":
                variable_lower_constraint = variable.get_lower_constraint()
                variable_upper_constraint = variable.get_upper_constraint()
                if variable_lower_constraint is not None:
                    convert_expression_identifiers_to_basic(
                        variable_lower_constraint)

                if variable_upper_constraint is not None:
                    convert_expression_identifiers_to_basic(
                        variable_upper_constraint)
                action_var = Action(variable_name, node_name,
                                    variable_lower_constraint,
                                    variable_upper_constraint)
                list_actions.append(action_var)

            elif var_option == "auxiliary":
                variable_assignment = variable.get_assignment()
                if variable_assignment is not None:
                    convert_expression_identifiers_to_basic(variable_assignment)
                aux_var = Auxiliary(variable_name, node_name,
                                    variable_assignment)
                list_auxiliaries.append(aux_var)

    for node in nodes:
        node_name = node.get_name()
        node_obj = node.get_objectives()
        for objective in node_obj:
            mdp_objective = MDPObjective(node_name, objective.get_expression())
            list_mdpobjectives.append(mdp_objective)

    mdp = MDP(list_states, list_actions, list_sizing, list_auxiliaries,
              list_mdpobjectives)
    return mdp


def check_all_variables(program_variables_dict):
    error_message = ""
    raised_error = False
    for node_name in program_variables_dict.keys():
        node_variables_dict = program_variables_dict[node_name]
        for variable_name in node_variables_dict.keys():
            variable = node_variables_dict[variable_name]
            var_option = variable.get_option()
            if var_option == "state":
                dynamic = variable.get_dynamic()
                initial = variable.get_initial_constraint()
                if dynamic is None or initial is None:
                    error_message += "\nERROR: State variable " + str(
                        variable_name) + " in node " + str(node_name) + \
                                     " is not properly defined : "

                    if dynamic is None:
                        error_message += "\n- the dynamic is missing"
                    if initial is None:
                        error_message += "\n- the initial condition is missing"

                    raised_error = True

            elif var_option == "action" or var_option == "sizing":
                lower_bound = variable.get_lower_constraint()
                upper_bound = variable.get_upper_constraint()
                if lower_bound is None or upper_bound is None:
                    error_message += "\nWarning: " + str(
                        var_option).title() + " variable " \
                                     + str(variable_name) + " in node " + str(
                        node_name) + " is not properly defined : "

                    if lower_bound is None:
                        error_message += "\n- A lower bound is missing"

                    if upper_bound is None:
                        error_message += "\n- An upper bound is missing"

            elif var_option == "auxiliary":
                assignment = variable.get_assignment()
                if assignment is None:
                    error_message += "\nERROR: Auxiliary variable " + str(
                        variable_name) + " in node " \
                                     + str(
                        node_name) + " is not properly defined : "
                    error_message += "\n- An assignment is missing"
                    raised_error = True
                circular_ref = check_circular_ref(variable, list())
                if circular_ref:
                    error_message += "\nERROR: Auxiliary variable " \
                                     + str(variable_name) \
                                     + "'s definition contains a " \
                                       "circular reference"
                    raised_error = True

    if raised_error:
        error_(error_message)
    else:
        print(error_message)


def check_circular_ref(variable: Variable, set_explored: list):
    list_aux = variable.get_dependencies()
    set_explored.append(variable)
    circular_ref = False
    for aux_var in list_aux:
        if aux_var in set_explored:
            circular_ref = True
        else:
            circular_ref = check_circular_ref(aux_var, set_explored)

        if circular_ref:
            break
    _ = set_explored.pop()
    return circular_ref


#
# Name checking functions
#


def match_dictionaries(dict1: dict, dict2: dict, dict3=None) -> None:
    """
    Match dictionaries find the intersection between
    the keys of two dictionaries returns nothing if set is empty
    and outputs an error otherwise
    INPUT:  dict1 -> dictionary
            dict2 -> dictionary
    OUTPUT: None, or error if fails
    """
    if dict3 is None:
        dict3 = {}
    dict1_set = set(dict1)
    dict2_set = set(dict2)
    dict3_set = set(dict3)
    inter_set = dict1_set.intersection(dict2_set, dict3_set)

    if len(inter_set) != 0:
        error_(
            "ERROR : some variables and parameters share the same name: "
            + str(inter_set))


def check_layer_repetition(nodes_list: list, hyperedges_list: list,
                           parent_nodes=None) -> None:
    if parent_nodes is None:
        parent_nodes = []
    check_names_repetitions(nodes_list + hyperedges_list + parent_nodes)

    for node in nodes_list:
        sub_nodes = node.get_sub_nodes()
        sub_hyperedges = node.get_sub_hyperedges()
        parent_nodes.append(node)
        check_layer_repetition(sub_nodes, sub_hyperedges, parent_nodes)
        parent_nodes.remove(node)


def check_names_repetitions(elements_list: list) -> None:
    """
    Checks if a node name is present twice in a list of nodes
    INPUT:  elements_list -> list of node objects
    OUTPUT: None, or error if fails
    """

    n = len(elements_list)
    i = 0

    for e in elements_list:
        name = e.get_name()

        if name == "T" or name == "t":
            error_('ERROR: Name "' + str(
                name) + '" is reserved for time, used at line ' + str(
                elements_list[i].get_line()))

        for k in range(i + 1, n):
            if name == elements_list[k].get_name() and name is not None:
                error_('ERROR: Redefinition error: "' + str(
                    name) + '" at line ' + str(elements_list[k].get_line()))
        i = i + 1


#
# End Name checking functions
#


#
# Link functions
#


def check_link_rl(hyperlink: Hyperedge, var_obj: dict) -> list:
    """
    check_link function : Takes program object and checks its links
    INPUT:  program -> Program object
    OUTPUT: list of input output pairs 
    """

    links = hyperlink.get_constraints()
    list_factor = []
    for link in links:
        if link.get_time_interval() or link.get_condition():
            error_(
                "ERROR: adding a time range is not"
                " allowed in MDP definition at line "
                + link.get_line())

        rhs = link.get_rhs()
        lhs = link.get_lhs()
        link_type = link.get_type()
        var_in_right = predicate_variables_in_expression(rhs, var_obj)
        var_in_left = predicate_variables_in_expression(lhs, var_obj)

        if link_type == "==":
            # Either auxiliary assignment of auxiliary or a dynamic definition
            if not var_in_left:
                error_("The left-hand-side should contain a variable at line "
                       + str(link.get_line()))
            variable, time_dep, node_name = check_lhs_equality(lhs, var_obj)
            var_option = variable.get_option()
            if var_option == "state":
                if time_dep != 2 and time_dep != 0:
                    error_("ERROR: can not assign a state at "
                           "another timestep than 't+1' at line "
                           + str(link.get_line()))
                if time_dep == 2:
                    check_rhs_equality_states(rhs, var_obj, variable, node_name)
                    try:
                        variable.set_dynamic(rhs)
                    except RedefinitionError:
                        error_("Error: redefinition of dynamic for variable "
                               + str(variable) +
                               " at line " + str(link.get_line()))
                if time_dep == 0 and var_in_right:
                    error_("ERROR: No variable can be used to "
                           "define an intial condition at line "
                           + str(link.get_line()))
                elif time_dep == 0:
                    try:
                        variable.set_initial_constraint(rhs)
                    except RedefinitionError:
                        error_(
                            "Error: redefinition of dynamic for variable "
                            + str(variable) +
                            " at line " + str(link.get_line()))

            elif var_option == "auxiliary":
                if time_dep != 1:
                    error_(
                        "ERROR : An auxiliary variable can only be "
                        "assigned at index t at line "
                        + str(link.get_line()))
                list_aux = check_rhs_equality_auxiliary(rhs, var_obj, variable,
                                                        node_name)
                try:
                    variable.set_assignment(rhs)
                    variable.set_dependencies(list_aux)
                except RedefinitionError:
                    error_("Error: redefinition of dynamic for variable " + str(
                        variable) +
                           " at line " + str(link.get_line()))
            else:
                error_(
                    "ERROR: assignments can only be used for "
                    "auxiliaries or states at line " + str(link.get_line()))
        elif link_type == "<=" or link_type == ">=":
            if var_in_right and var_in_left:
                error_(
                    "ERROR : the left and right hand-side of an "
                    "inequality can not contain variables at line "
                    + str(link.get_line()))
            elif var_in_right:
                variable, _ = check_variable_inequality(rhs, var_obj)
                try:
                    if link_type == "<=":
                        variable.set_lower_constraint(lhs)
                    if link_type == ">=":
                        variable.set_upper_constraint(lhs)
                except RedefinitionError:
                    error_(
                        "Error: redefinition of lower-bound for variable "
                        + str(variable) +
                        " at line " + str(link.get_line()))

            elif var_in_left:
                variable, _ = check_variable_inequality(lhs, var_obj)
                try:
                    if link_type == "<=":
                        variable.set_upper_constraint(rhs)
                    if link_type == ">=":
                        variable.set_lower_constraint(rhs)
                except RedefinitionError:
                    error_(
                        "Error: redefinition of lower-bound for variable "
                        + str(variable) +
                        " at line " + str(link.get_line()))

            else:
                error_("ERROR: No variable in constraint at line " + str(
                    link.get_line()))
    return list_factor


#
# End link functions
#

#
# Node checking functions
#


def check_variables_type_sizes(dictionary_var: dict, dictionary_param: dict):
    timehorizon = dictionary_param["T"].get_value()[0]
    for node_name in dictionary_var.keys():
        for variable_name in dictionary_var[node_name].keys():
            var = dictionary_var[node_name][variable_name]
            v_size = var.get_size()
            v_type = var.get_option()

            if v_type == "auxiliary" and v_size != timehorizon:
                error_(
                    "ERROR: auxiliary variables must have a size of T at line "
                    + str(var.get_line()))

            elif v_type == "state" and v_size != timehorizon:
                error_(
                    "ERROR: state variables must have a size of T at line "
                    + str(var.get_line()))

            elif v_type == "sizing" and v_size != 1:
                error_(
                    "ERROR: sizing variables must have a size of 1 at line "
                    + str(var.get_line()))


def set_size_variables(dictionary_var: dict, dictionary_param: dict, index: int,
                       nested_nodes_variables) -> int:
    """
    Initializes the index of variable object inside a dictionary
    """
    start_index = index
    for k in dictionary_var.keys():
        var = dictionary_var[k]
        identifier = var.get_identifier()
        identifier.set_size(dictionary_param)
        child_assignment_identifier = var.get_child_assignment()
        if child_assignment_identifier:
            child_node_name = child_assignment_identifier.get_node_name()
            child_var_name = child_assignment_identifier.get_name()
            child_expression = child_assignment_identifier.get_expression()
            if child_expression is None:
                child_declared_size = 1
            else:
                child_declared_size = child_expression.evaluate_expression(
                    dictionary_param)
            if child_assignment_identifier.get_type() != identifier.get_type():
                error_(
                    "ERROR: assigning variables of "
                    "different types is not allowed at line : "
                    + str(identifier.get_line()))
            if not child_node_name:
                error_(
                    "ERROR: node name is required for assignments at line : "
                    + str(identifier.get_line()))

            if child_node_name in nested_nodes_variables and child_var_name in \
                    nested_nodes_variables[child_node_name]:

                child_variable = nested_nodes_variables[child_node_name][
                    child_var_name]
                child_id = child_variable.get_identifier()
                identifier.set_index(child_id.get_index())
                if identifier.get_size() != child_id.get_size() or \
                        identifier.get_size() != child_declared_size:
                    error_(
                        "ERROR: unmatching size in assignment "
                        "of variable at line : " + str(identifier.get_line()))
            else:
                error_('ERROR: Unknown identifier ' + str(
                    child_assignment_identifier) + " not defined at line " +
                       str(identifier.get_line()))
        else:
            start_index = identifier.set_index(start_index)

    return start_index


def check_expressions_dependancy_rl(node: Node, variables_dict: dict):
    constraints = node.get_constraints()
    objectives = node.get_objectives()
    node_name = node.get_name()
    for constraint in constraints:
        if constraint.get_time_interval() or constraint.get_condition():
            error_(
                "ERROR: adding a time range "
                "is not allowed in MDP definition at line "
                + constraint.get_line())

        rhs = constraint.get_rhs()
        lhs = constraint.get_lhs()
        constr_type = constraint.get_type()
        no_sum_in_expression(rhs)
        no_sum_in_expression(lhs)

        var_in_right = predicate_variables_in_expression(rhs, variables_dict)
        var_in_left = predicate_variables_in_expression(lhs, variables_dict)

        if constr_type == "==":
            # Either auxiliary assignment of auxiliary or a dynamic definition

            variable, time_dep, _ = check_lhs_equality(lhs, variables_dict)
            var_option = variable.get_option()
            if var_option == "state":
                if time_dep != 2 and time_dep != 0:
                    error_(
                        "ERROR: can not assign a state "
                        "at another timestep than 't+1' at line "
                        + str(constraint.get_line()))
                if time_dep == 2:
                    check_rhs_equality_states(rhs, variables_dict, variable,
                                              var_node_name=node_name)
                    try:
                        variable.set_dynamic(rhs)
                    except RedefinitionError:
                        error_(
                            "Error: redefinition of dynamic for variable "
                            + str(variable) +
                            " at line " + str(constraint.get_line()))
                if time_dep == 0 and var_in_right:
                    error_(
                        "ERROR: No variable can be used to "
                        "define an intial condition at line "
                        + str(constraint.get_line()))
                elif time_dep == 0:
                    try:
                        variable.set_initial_constraint(rhs)
                    except RedefinitionError:
                        error_(
                            "Error: redefinition of dynamic for variable "
                            + str(variable) +
                            " at line " + str(constraint.get_line()))
            elif var_option == "auxiliary":
                if time_dep != 1:
                    error_(
                        "ERROR: can not assign an auxiliary "
                        "at another timestep than 't' at line "
                        + str(constraint.get_line()))
                list_aux = check_rhs_equality_auxiliary(rhs, variables_dict,
                                                        variable)
                try:
                    variable.set_assignment(rhs)
                    variable.set_dependencies(list_aux)
                except RedefinitionError:
                    error_("Error: redefinition of dynamic for variable " + str(
                        variable) +
                           " at line " + str(constraint.get_line()))
            else:
                error_(
                    "ERROR: assignments can only be used "
                    "for auxiliaries or states at line "
                    + str(constraint.get_line()))

        elif constr_type == "<=" or constr_type == ">=":
            if var_in_right and var_in_left:
                error_(
                    "ERROR : the left and right hand-side "
                    "of an inequality can not contain variables at line "
                    + str(constraint.get_line()))
            elif var_in_right:
                variable, _ = check_variable_inequality(rhs, variables_dict)
                try:
                    if constr_type == "<=":
                        variable.set_lower_constraint(lhs)
                    if constr_type == ">=":
                        variable.set_upper_constraint(lhs)
                except RedefinitionError:
                    error_(
                        "Error: redefinition of lower-bound for variable "
                        + str(variable) +
                        " at line " + str(constraint.get_line()))

            elif var_in_left:
                variable, _ = check_variable_inequality(lhs, variables_dict)
                try:
                    if constr_type == "<=":
                        variable.set_upper_constraint(rhs)
                    if constr_type == ">=":
                        variable.set_lower_constraint(rhs)
                except RedefinitionError:
                    error_(
                        "Error: redefinition of lower-bound for variable "
                        + str(variable) +
                        " at line " + str(constraint.get_line()))

            else:
                error_("ERROR: No variable in constraint at line " + str(
                    constraint.get_line()))

    for objective in objectives:
        expr = objective.get_expression()
        no_sum_in_expression(expr)
        check_objective(expr, variables_dict)


def no_sum_in_expression(expression: Expression):
    leaves = expression.get_leafs()
    for leaf in leaves:
        leaf_type = leaf.get_type()
        seed = leaf.get_name()
        if leaf_type == "sum":
            error_(
                "ERROR: sum operators are not allowed "
                "for MDP definition at line " + str(leaf.get_line()))
        elif type(seed) == Identifier:
            identifier: Identifier = seed
            identifier_type = identifier.get_type()
            if identifier_type == "assign":
                identifier_expression = identifier.get_expression()
                no_sum_in_expression(identifier_expression)


def check_objective(expression: Expression, var_dict: dict):
    leaves = expression.get_leafs()
    list_var = []

    for leaf in leaves:

        seed = leaf.get_name()
        if type(seed) == Identifier:
            l_name = seed.get_name()

            if l_name in var_dict:
                variable_leaf = var_dict[l_name]
                time_dep = check_time_dependancy(seed)
                if time_dep != 1 and time_dep != -1:
                    error_("ERROR: term " + str(seed)
                           + " at that timestep is not allowed "
                             "for the definition of "
                             "auxiliaries at line " + str(seed.get_line()))

                list_var.append(variable_leaf)

    return list_var


def check_rhs_equality_auxiliary(rhs: Expression, var_dict: dict,
                                 variable: Variable, var_node_name=""):
    leaves = rhs.get_leafs()
    list_aux = []
    var_id = variable.get_identifier()
    var_name = var_id.get_name()

    for leaf in leaves:

        seed = leaf.get_name()
        node_var = var_dict

        if type(seed) == Identifier:
            identifier: Identifier = seed
            id_name = identifier.get_name()
            node_name = identifier.get_node_name()
            if node_name in var_dict:
                node_var = var_dict[node_name]

            if id_name in node_var:
                variable_leaf = node_var[id_name]
                option = variable_leaf.get_option()
                time_dep = check_time_dependancy(identifier)
                if time_dep != 1 and time_dep != -1:
                    error_("ERROR: term " + str(seed)
                           + " at that timestep is not allowed "
                             "for the definition of auxiliaries at line "
                           + str(seed.get_line()))

                if option == "auxiliary":
                    if id_name == var_name and node_name == var_node_name:
                        error_(
                            "ERROR: left-hand-side and "
                            "right-hand-side contain the same term at line "
                            + str(seed.get_line()))

                    list_aux.append(variable_leaf)

    return list_aux


def check_rhs_equality_states(rhs: Expression, var_dict: dict, variable,
                              var_node_name=""):
    leaves = rhs.get_leafs()

    var_id: Identifier = variable.get_name()
    var_name = var_id.get_name()

    time_dep = None
    is_state = False
    is_action = False
    previous_step = False

    for leaf in leaves:

        seed = leaf.get_name()
        node_dict = var_dict
        if type(seed) == Identifier:
            identifier: Identifier = seed
            identifier_name = identifier.get_name()
            node_name = identifier.get_node_name()
            if node_name in var_dict:
                node_dict = var_dict[node_name]

            if identifier_name in node_dict:
                variable = node_dict[identifier_name]
                option = variable.get_option()

                if option == "action":
                    is_action = True

                elif option == "state":
                    is_state = True
                    if identifier_name == var_name and \
                            node_name == var_node_name:
                        previous_step = True

                time_dep = check_time_dependancy(identifier)

                if time_dep != 1:
                    error_("ERROR: term " + str(seed)
                           + " at that timestep is not "
                             "allowed for the definition of "
                             "dynamics at line " + str(seed.get_line()))

    if is_action is not True or previous_step is not True:
        error_message = "ERROR : dynamic at line: " + str(
            rhs.get_line()) + " is ill-defined :"

        if is_action is not True:
            error_message += "\n- an action is missing"
        if previous_step is not True:
            error_message += "\n- the previous state is missing"
        if is_state is not True:
            error_message += "\n- No state is used"
        error_(error_message)

    return variable, time_dep


def check_lhs_equality(lhs: Expression, var_dict: dict):
    leaves = lhs.get_leafs()
    variable = None
    time_dep = None
    node_name = ""
    if len(leaves) != 1:
        error_(
            "ERROR: not only a single left-hand-side "
            "variable is allowed in constraints "
            "at line " + str(lhs.get_line()))

    for leaf in leaves:
        seed = leaf.get_name()
        if type(seed) == Identifier:
            identifier: Identifier = seed
            identifier_name = identifier.get_name()
            identifier_node_name = identifier.get_node_name()
            if identifier_node_name in var_dict and \
                    identifier_name in var_dict[identifier_node_name]:
                variable = var_dict[identifier_node_name][identifier_name]
                time_dep = check_time_dependancy(identifier)

            else:
                error_(
                    "ERROR: not allowed to use anything "
                    "but variables in the left-hand-side "
                    "of equality constraints at line " + str(lhs.get_line()))

        else:
            error_(
                "ERROR: not allowed to use anything "
                "but variables in the left-hand-side "
                "of equality constraints at line " + str(lhs.get_line()))

    return variable, time_dep, node_name


def check_variable_inequality(expr: Expression, var_dict: dict):
    leaves = expr.get_leafs()
    variable = None
    time_dep = None

    if len(leaves) != 1:
        error_("ERROR: only a single variable is allowed in constraints "
               "at line " + str(expr.get_line()))

    for leaf in leaves:

        seed = leaf.get_name()
        node_dict = var_dict

        if type(seed) == Identifier:
            identifier: Identifier = seed
            node_name = identifier.get_node_name()
            id_name = identifier.get_name()
            if node_name in var_dict:
                node_dict = var_dict[node_name]

            if id_name in node_dict:
                variable = node_dict[id_name]
                option = variable.get_option()
                if option != "sizing" and option != "action":
                    error_(
                        "ERROR: only action and sizing variables "
                        "can be concerned with inequality constraints")
                time_dep = check_time_dependancy(identifier)
                if option == "sizing" and time_dep != -1:
                    error_("Error: sizing variables " + str(
                        leaf) + " must be not indexed at line " + str(
                        leaf.get_line()))
                if option == "action" and time_dep != 1:
                    error_("Error: action variables " + str(
                        leaf) + " must be indexed at t at line " + str(
                        leaf.get_line()))

            else:
                error_(
                    "ERROR: not allowed to use anything "
                    "but variables in the inequality "
                    "constraints at line " + str(expr.get_line()))

    return variable, time_dep


def check_time_dependancy(seed):
    seed_type = seed.get_type()
    return_value = 0
    if seed_type == "basic":
        return_value = -1

    elif seed_type == "assign":
        expr = seed.get_expression()
        expr_type = expr.get_type()
        if expr_type == "literal":
            identifier = expr.get_name()
            if type(identifier) == int or type(identifier) == float:
                value = identifier
                if value != 0:
                    error_(
                        "ERROR : constants other than 0 are "
                        "not allowed as in bracket expression at line "
                        + str(expr.get_line()))

            elif identifier.get_name() != "t":

                error_(
                    "ERROR : the in bracket expression can not be a parameter"
                    + str(identifier.get_name()))

            else:
                return_value = 1

        elif expr_type == "+":
            child_1, child_2 = expr.get_children()
            child_1_type = child_1.get_type()
            child_2_type = child_2.get_type()
            child_1_identifier = child_1.get_name()
            child_2_identifier = child_2.get_name()
            rhs_number = False
            lhs_number = False

            if child_1_type != "literal" or child_2_type != "literal":
                error_('ERROR: the assignment ' + str(
                    expr) + ' is not allowed at line ' + str(expr.get_line()))

            if type(child_1_identifier) == int or type(
                    child_1_identifier) == float:
                value = child_1_identifier
                if value != 1:
                    error_(
                        "ERROR : constants other than 0 are not "
                        "allowed as in bracket expression at line "
                        + str(expr.get_line()))
                lhs_number = True

            elif child_1_identifier.get_name() != "t":

                error_(
                    "ERROR : the in bracket expression can not be a parameter"
                    + str(child_1_identifier.get_name()))

            if (type(child_2_identifier) == int or type(
                    child_2_identifier) == float) and not rhs_number:
                value = child_2_identifier
                if value != 1:
                    error_(
                        "ERROR : constants other than 0 are not "
                        "allowed as in bracket expression at line "
                        + str(expr.get_line()))
                rhs_number = True

            elif child_2_identifier.get_name() != "t":

                error_(
                    "ERROR : the in bracket expression can not be a parameter"
                    + str(child_2_identifier.get_name()))

            if (lhs_number or rhs_number) and not (lhs_number and rhs_number):

                return_value = 2

            else:
                error_('ERROR: the assignment ' + str(
                    expr) + ' is not allowed at line ' + str(expr.get_line()))

        else:
            error_('ERROR: the assignment ' + str(
                expr) + ' is not allowed at line ' + str(expr.get_line()))
    return return_value


def predicate_variables_in_expression(expression: Expression,
                                      variables: dict) -> bool:
    """
    variables_in_expression function : returns true if expression
    contains variables and false otherwise
    INPUT:  expression -> expression object
            variables -> dictionary of <name,identifier> objects
            parameters -> dictionary of <name,array> objects 
            check_in -> check for errors in identifier 's assigned expression
    OUTPUT: bool -> boolean value if expression contains variable
    """

    leaves = expression.get_leafs()
    is_variable: bool = False

    for leaf in leaves:
        seed = leaf.get_name()
        leaf_type = leaf.get_type()
        if type(seed) == Identifier:
            identifier: Identifier = seed
            identifier_name = identifier.get_name()
            identifier_node_name = identifier.get_node_name()
            if identifier_node_name in variables and identifier_name in \
                    variables[identifier_node_name]:
                is_variable = True
                break

        elif leaf_type == 'sum':

            children_of_sum = leaf.get_children()
            for child in children_of_sum:

                is_child_var = predicate_variables_in_expression(child,
                                                                 variables)
                if is_child_var is True:
                    is_variable = True
                    break

    return is_variable


def parameter_evaluation(parameter_dict: dict, definitions: dict):
    """
    parameter_evaluation function : evaluates a list of parameter objects
    INPUT:  n_parameters -> list of parameters objects
            definitions -> dictionary of definitions <name,array>
    OUTPUT: definitions -> dictionary of definitions <name,array>
    """
    input_definition = definitions.copy()
    for parameter_name in parameter_dict.keys():
        parameter = parameter_dict[parameter_name]
        e = parameter.get_expression()
        if e is not None:

            value = e.evaluate_expression(input_definition)
            value = [value]
        else:

            value = evaluate_table(parameter.get_vector(), input_definition)

        parameter.set_value(value)
        input_definition[parameter_name] = parameter


def evaluate_table(list_values: list, definitions: dict) -> list:
    """
    evaluate_table function : evaluates a list of expression objects
    INPUT:  list_values -> list of expression objects
            definitions -> dictionary of definitions <name,value>
    OUTPUT: list <float>
    """

    all_values: list = []
    for value in list_values:

        value_i = value.get_name()
        if type(value_i) == Identifier:

            type_val = value_i.get_type()
            id_name = value_i.get_name()
            if not (id_name in definitions):
                error_('Undefined parameter : ' + str(value_i))
            values = definitions[id_name]
            if type_val == "basic" and len(values) == 1:

                value_i = values[0]
            elif type_val == "basic" and len(values) > 1:

                error_('Parameter not properly defined : ' + str(value_i))
            elif type_val == "assign":

                inner_expr = value_i.get_expression()
                index = inner_expr.evaluate_expression(definitions)
                if type(index) == float:

                    if index.is_integer() is False:
                        error_("Error: an index is a float: " + str(value_i))
                    index = int(round(index))

                if index < 0 or len(values) <= index:
                    error_('Parameter does not exit at this index : ' + str(
                        value_i))
                value_i = values[index]
        all_values.append(value_i)
    return all_values

#
# END Expression FUNCTIONS
#



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .gboml_output import generate_json, generate_list_values_tuple, write_csv



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""Gurobi Solver file, contains the interface to Gurobi solver .

Takes the matrix A, the vector b and vector c as input of the problem
    min : c^T * X s.t. A * X <= b
and passes it to the gurobi solver.

  Typical usage example:

   solution, objective, status, solver_info = gurobi_solver(matrix_a, vector_b,
                                                            vector_c,
                                                            objective_offset,
                                                            name_tuples)
   print("the solution is "+str(solution))
   print("the objective found : "+str(objective))
"""

import numpy as np
from scipy.sparse import coo_matrix
from gboml.compiler.utils import flat_nested_list_to_two_level
import os


def gurobi_solver(matrix_a: coo_matrix, vector_b: np.ndarray,
                  vector_c: np.ndarray,
                  objective_offset: float,
                  name_tuples: dict,
                  opt_file: str = None) -> tuple:
    """gurobi_solver

        takes as input the matrix A, the vectors b and c. It returns
        the solution of the problem : min c^T * x s.t. A * x <= b found
        by the gurobi solver

        Args:
            A -> coo_matrix of constraints
            b -> np.ndarray of independent terms of each constraint
            c -> np.ndarray of objective vector
            objective_offset -> float of the objective offset
            name_tuples -> dictionary of <node_name variables> used to get
                           the type
            opt_file -> optimization parameters file

        Returns:
            solution -> np.ndarray of the flat solution
            objective -> float of the objective value
            status -> solution status
            solver_info -> dictionary of solver information
            constraints_additional_information -> dict of additional
                                                  information concerning
                                                  constraints
            variables_additional_information -> dict of additional information
                                                concerning variables

    """

    try:
        import gurobipy as grbp
        from gurobipy import GRB

    except ImportError:
        print("Warning: Did not find the gurobipy package")
        exit(0)

    if opt_file is None:
        opt_file = 'gboml/solver_api/gurobi.opt'

    solution = None
    objective = None

    # Fix shapes and types of input
    matrix_a = matrix_a.astype(float)
    _, n = np.shape(matrix_a)
    b = vector_b.reshape(-1)
    vector_c = vector_c.reshape(-1)

    # Build Gurobi model
    model = grbp.Model()
    x = model.addMVar(shape=n, lb=-float('inf'), ub=float('inf'),
                      vtype=GRB.CONTINUOUS, name="x")
    model.addMConstr(matrix_a, x, '<', b)
    model.setObjective(vector_c @ x + objective_offset, GRB.MINIMIZE)

    flat_name_tuples = flat_nested_list_to_two_level(name_tuples)

    for index, _, var_type, var_size in flat_name_tuples:

        if var_type == "integer":
            x[index:index + var_size].vtype = GRB.INTEGER
        if var_type == "binary":
            x[index:index + var_size].vtype = GRB.BINARY

    # Gather and retrieve solver information
    solver_info = dict()
    solver_info["name"] = "gurobi"
    print("\nReading Gurobi options from file gurobi.opt")
    option_info = {}

    try:

        with open(opt_file, 'r') as optfile:
            lines = optfile.readlines()
    except IOError:

        print("Options file not found")
    else:

        for line in lines:

            line = line.strip()
            option = line.split(" ", 1)
            if option[0] != "":

                parinfo = model.getParamInfo(option[0])
                if parinfo:

                    if len(option) == 2:

                        key = option[0]
                        try:

                            value = parinfo[1](option[1])
                        except ValueError as e:

                            print("Skipping option \'%s\' with invalid "
                                  "given value \'%s\' (expected %s)"
                                  % (option[0], option[1], parinfo[1]))
                        else:

                            model.setParam(key, value)
                            option_info[key] = value
                    else:

                        print("Skipping option \'%s\' with no given value"
                              % option[0])
                else:

                    print("Skipping unknown option \'%s\'" % option[0])

    solver_info["options"] = option_info
    print("")

    # Solve model and return solution
    try:

        model.optimize()
        status_code = model.getAttr("Status")
        solver_info["status"] = status_code
        solution = None
        objective = None
        if status_code == 2:

            status = "optimal"
            solution = x.X
            objective = model.getObjective().getValue()
        elif status_code == 3:

            status = "infeasible"
            objective = float('inf')
        elif status_code == 5:

            status = "unbounded"
            objective = float('-inf')
        elif status_code == 13:

            status = "feasible"
            solution = x.X
            objective = model.getObjective().getValue()
        else:

            status = "unknown"
    except RuntimeError as e:

        print(e)
        status = "error"

    constraints_additional_information = dict()
    variables_additional_information = dict()
    attributes_to_retrieve_constraints = ["Pi", "Slack",
                                          "CBasis", "SARHSLow", "SARHSUp"]
    attributes_to_retrieve_variables = ["RC", "VBasis", "SAObjLow", "SAObjUp",
                                        "SALBLow", "SALBUp", "SAUBLow",
                                        "SAUBUp"]

    for attribute in attributes_to_retrieve_constraints:
        try:
            constraints_additional_information[attribute] = \
                model.getAttr(attribute, model.getConstrs())
        except grbp.GurobiError:
            print("Warning : Unable to retrieve ", attribute,
                  " information for constraints")

    for attribute in attributes_to_retrieve_variables:
        try:
            variables_additional_information[attribute] = \
                model.getAttr(attribute, model.getVars())
        except grbp.GurobiError:
            print("Warning : Unable to retrieve ", attribute,
                  " information for variables")

    return solution, objective, status, solver_info, \
           constraints_additional_information, \
           variables_additional_information



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .time_obj import Time, TimeInterval
from .constraint import Constraint
from .expression import Expression
from .identifier import Identifier
from .link import Attribute, Hyperedge
from .parameter import Parameter
from .program import Program 
from .variable import Variable
from .objective import Objective
from .node import Node
from .condition import Condition
from .factor import Factorize
from .mdp import MDP, State, Action, Auxiliary, Sizing, MDPObjective

__all__ = ["Constraint", "Expression", "Identifier", "Node", "Parameter",
           "Program", "Time", "TimeInterval", "Variable", "Objective",
           "Condition", "Factorize", "Attribute", "Hyperedge", "MDP", "State",
           "Sizing", "Action", "Auxiliary", "MDPObjective"]



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from gboml.compiler.utils import error_, turn_to_dict


class Node:
    """
    Node object is composed of: 
    - list of constraints
    - list of parameters
    - list of objectives 
    - list of variables
    - list of links related to the node
    - variable matrix (each column is an identifier of one variables)
    - triplet [array A, sign , b] for constraints
    - list of objective arrays
    - counter for number of constraints
    - all parameters dictionary [name, [values]]
    """

    def __init__(self, name, line=0):

        self.name = name
        self.constraints = []
        self.variables = []
        self.parameters = []
        self.objectives = []
        self.line = line
        self.links = []
        self.v_matrix = None
        self.c_triplet_list = []
        self.objective_list = []
        self.nb_constraint_matrix = 0
        self.nb_objective_matrix = 0
        self.param_dict = None
        self.constr_factors = []
        self.obj_factors = []
        self.nodes = []
        self.hyperedges = []
        self.expression = []
        self.parameters_changes = []
        self.variables_changes = []
        self.dict_sub_nodes_edges = {}
        self.objectives_data = {}
        self.constraints_data = {}

    def __str__(self):

        string = '[' + str(self.name) + ' , '
        string += str(self.parameters) + ' , '
        string += str(self.variables) + ' , '
        string += str(self.constraints) + ' , '
        string += str(self.objectives) + ']'

        return string

    def set_line(self, line):

        self.line = line

    def get_line(self):

        return self.line

    def get_sub_nodes(self):
        return self.nodes

    def set_sub_nodes(self, sub_nodes):
        self.nodes = sub_nodes

    def add_sub_node(self, node):
        self.nodes.append(node)

    def get_sub_hyperedges(self):
        return self.hyperedges

    def set_objectives_data(self, obj_data):
        self.objectives_data = obj_data

    def get_objectives_data(self):
        return self.objectives_data

    def set_constraints_data(self, constr_data):
        self.constraints_data = constr_data

    def get_constraints_data(self):
        return self.constraints_data

    def set_sub_hyperedges(self, sub_hyperedges):
        self.hyperedges = sub_hyperedges

    def set_variables_changes(self, changes):
        self.variables_changes = changes

    def add_variable_change(self, change):
        self.variables_changes.append(change)

    def get_variables_changes(self):
        return self.variables_changes

    def set_parameters_changes(self, changes):
        self.parameters_changes = changes

    def add_parameter_change(self, change):
        self.parameters_changes.append(change)

    def get_parameters_changes(self):
        return self.parameters_changes

    def set_objective_factors(self, fact_list):

        self.obj_factors = fact_list

    def set_constraint_factors(self, fact_list):

        self.constr_factors = fact_list

    def update_internal_dict(self):

        self.dict_sub_nodes_edges = turn_to_dict(self.get_sub_nodes()
                                                 + self.get_sub_hyperedges())

    def get_internal_dict(self):
        return self.dict_sub_nodes_edges

    def get_objective_factors(self):

        return self.obj_factors

    def free_factors_objectives(self):
        self.obj_factors = None

    def free_factors_constraints(self):
        for factor in self.constr_factors:
            factor.sparse = None
            factor.independent_terms = None
        self.constr_factors = None

    def get_constraint_factors(self):

        return self.constr_factors

    def set_parameter_dict(self, param):

        param = param.copy()
        if "global" in param:
            param.pop("global")
        if "GLOBAL" in param:
            param.pop("GLOBAL")
        if "T" in param:
            param.pop("T")
        self.param_dict = param

    def add_link(self, link):

        self.links.append(link)

    def get_links(self):

        return self.links

    def set_expressions(self, list_expression):
        self.expression = list_expression

    def get_expressions(self):
        return self.expression

    def set_constraints(self, cons):

        self.constraints = cons

    def set_variables(self, var):

        self.variables = var

    def set_parameters(self, para):

        self.parameters = para

    def set_objectives(self, obj):

        self.objectives = obj

    def get_name(self):

        return self.name

    def rename(self, new_name):
        self.name = new_name

    def get_constraints(self):

        return self.constraints

    def remove_constraint(self, constraint):
        self.constraints.remove(constraint)

    def remove_objective(self, objective):
        self.objectives.remove(objective)

    def get_number_constraints(self):

        return len(self.constraints)

    def get_number_expanded_constraints(self, with_sub_nodes_and_edges=False):

        total_number_of_constraint = self.nb_constraint_matrix

        if with_sub_nodes_and_edges:
            for sub_node in self.get_sub_nodes():
                total_number_of_constraint += \
                    sub_node.get_number_expanded_constraints(True)

            for sub_hyperedge in self.get_sub_hyperedges():
                total_number_of_constraint += \
                    sub_hyperedge.get_number_expanded_constraints()

        return total_number_of_constraint

    def get_parameter_dict(self):

        return self.param_dict

    def get_variables(self):

        return self.variables

    def get_number_variables(self):

        return len(self.variables)

    def get_variable_names(self):

        names = []
        for var in self.variables:
            names.append(var.get_name().get_name())

        return names

    def get_parameters(self):

        return self.parameters

    def get_number_parameters(self):

        return len(self.parameters)

    def get_objectives(self):

        return self.objectives

    def get_number_objectives(self):

        return len(self.objectives)

    def get_number_expanded_objectives(self):

        return self.nb_objective_matrix

    def set_variable_matrix(self, var_matrix):

        self.v_matrix = var_matrix

    def get_variable_matrix(self):

        return self.v_matrix

    def set_constraints_matrix(self, list_matrix):
        self.nb_constraint_matrix += len(list_matrix)
        self.c_triplet_list = list_matrix

    def add_constraints_matrix(self, c_matrix):

        self.nb_constraint_matrix += 1
        self.c_triplet_list.append(c_matrix)

    def get_constraints_matrix(self):

        return self.c_triplet_list

    def set_objective_matrix(self, o):

        length = 0
        for obj_index, tuple_obj in o:
            length += len(tuple_obj)
        self.nb_objective_matrix = length
        self.objective_list = o

    def add_objective_matrix(self, o):

        self.objective_list.append(o)

    def get_objective_list(self):

        return self.objective_list

    def get_nb_constraints_matrix(self):

        return self.nb_constraint_matrix

    def get_dictionary_variables(self, get_type="all", get_id=True):

        variables = self.variables
        all_variables = {}
        reserved_names = ["t", "T"]
        for var in variables:

            v_type = var.get_type()
            if get_type == "external" and v_type == "internal":
                continue
            if get_type == "internal" and v_type == "external":
                continue
            identifier = var.get_identifier()
            name = identifier.get_name()
            if name in reserved_names:
                error_("Semantic error, variable named " + str(name) +
                       " is not allowed at line " + str(var.get_line()))
            if name not in all_variables:

                if get_id:
                    all_variables[name] = identifier
                else:
                    all_variables[name] = var
            else:

                error_("Semantic error, redefinition of variable " + str(name)
                       + " at line " + str(var.get_line()))

        return all_variables

    def get_dictionary_parameters(self):

        parameters = self.parameters
        all_parameters = dict()
        reserved_names = ["t", "T"]
        for param in parameters:

            name = param.get_name()
            if name in reserved_names:
                error_(
                    "Semantic error, variable named " + str(name)
                    + " is not allowed at line " + str(param.get_line()))
            if name not in all_parameters:

                all_parameters[name] = param
            else:

                error_("Semantic error, redefinition of variable " + str(name)
                       + " at line " + str(param.get_line()))

        return all_parameters

    def get_dictionary_expressions(self):

        expressions = self.expression
        all_expressions = dict()
        reserved_names = ["t", "T"]
        for name, expr, line in expressions:

            if name in reserved_names:
                error_("Semantic error, expression named " + str(name) +
                       " is not allowed at line " + str(line))
            if name not in all_expressions:

                all_expressions[name] = expr
            else:

                error_("Semantic error, redefinition of expression "
                       + str(name) + " at line " + str(line))

        return all_expressions



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""GBOML compiler file

Defines functions to extract the information of a GBOML file,
check and translate it to another structure.

  Typical usage example:

  compile_gboml(gboml_file)
  where:
    gboml_file is the file we want to compile

"""

from .gboml_lexer import tokenize_file
from .gboml_parser import parse_file
from .gboml_semantic import semantic, check_mdp, convert_to_mdp, \
    check_program_linearity, factorize_program
from .gboml_matrix_generation import matrix_generation_a_b, \
    matrix_generation_c, extend_factor, extend_factor_on_multiple_processes
from .utils import move_to_directory

import sys
import os


def compile_gboml(input_file: str, log: bool = False,
                  lex: bool = False, parse: bool = False,
                  nb_processes: int = 1) -> tuple:
    """compile_gboml

        takes as input a filename and converts to the matrix
        representation of the problem and a program object
        (abstract syntax tree)

        Args:
            input_file -> string containing the input file
            log -> boolean predicate of should the output log be saved
                   in the file
            lex -> boolean predicate of printing the different tokens
                   in the file
            parse -> boolean predicate of printing the abstract
                     syntax tree generated from the file
            nb_processes -> number of processes (workers) for
                            the model extension

        Returns:
             program -> program object
             matrix_a -> Constraint sparse matrix
             vector_b -> Vector of independent terms for each constraint
             vector_c -> objective sparse matrix
             indep_terms_c -> vector of independent terms of each row in
                              the objective sparse matrix
             T -> Time horizon
             name_tuples -> Mapping to convert the flat x solution
                            to the original graph structure

    """

    curr_dir, filename = move_to_directory(input_file)

    if log is True:
        filename_split = filename.rsplit('.', 1)
        logfile = filename_split[0]
        f = open(logfile + ".out", 'w')
        sys.stdout = f

    if lex is True:
        tokenize_file(filename)

    ast = parse_file(filename)
    if parse is True:
        print(ast.to_string())

    program, program_variables_dict, definitions = semantic(ast)
    check_program_linearity(program, program_variables_dict, definitions)
    factorize_program(program, program_variables_dict, definitions)
    if nb_processes > 1:
        extend_factor_on_multiple_processes(program, definitions, nb_processes)
    else:
        extend_factor(program, definitions)

    matrix_a, vector_b = matrix_generation_a_b(program)
    vector_c, indep_terms_c = matrix_generation_c(program)
    program.free_factors_objectives()

    time_horizon = program.get_time().get_value()
    os.chdir(curr_dir)
    return program, matrix_a, vector_b, vector_c, indep_terms_c, \
           time_horizon, program.get_tuple_name()


def compile_gboml_mdp(input_file: str):
    """compile_gboml_mdp

        takes as input a filename and converts to the mdp
        representation of the problem

        Args:
            input_file -> string containing the input file

        Returns:
             mdp -> MDP object containing the file information

    """

    curr_dir, filename = move_to_directory(input_file)
    ast = parse_file(filename)
    program, program_variables_dict, definitions = semantic(ast)
    check_mdp(program, program_variables_dict, definitions)
    mdp = convert_to_mdp(program, program_variables_dict)
    os.chdir(curr_dir)

    return mdp



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""
    Python class interface for Argonne National Laboratory (USA)'s Decomposition
     of Structured Programs (DSP) implemented by Miftari B and Berger Mathias
     from Uliege (Belgium).
    This interface provides a basic python mapping for all the deterministic
    functions present in DspCInterface.h

    example :

        from DSPpy import DSPpy
        dsp = DSPpy()
        pointer_to_env = dsp.createEnv()
        dsp.freeEnv(pointer_to_env)

"""

from ctypes import cdll, c_void_p, c_int, c_double, c_char, POINTER, byref
from ctypes.util import find_library


class DSPpy:
    """
    DSPpy is the Python class interface with the shared library
    Notes : - the shared library (libDsp.so on Linux or libDsp.dylib on OS X)
              should be on the library path
            - this implementation was not tested on Windows as DSP itself was
              not tested on Windows
            - For DSP installation steps, please refer to:
              https://github.com/Argonne-National-Laboratory/DSP

    """

    def __init__(self, path=None):
        """__init__

        Loads the shared DSP library and returns a DSPpy object which contains
        an interface to the different functions defined in DSPCInterface.h

        Args:
            path(str) : explicit path to the DSP library (if none,
                        the path is searched on the default library path)

        Returns:
            DSPpy object

        """
        if path is None:
            path = find_library("DSP")

        self.libDSP = cdll.LoadLibrary(path)

    def createEnv(self) -> c_void_p:
        """createEnv

        Initializes a C++ DSP environment and returns a ctype pointer
        to the environnement

        Args:

        Returns:
            pointer_to_env (c_void_p) : ctypes pointer to a DSP environment

        """
        self.libDSP.createEnv.restype = c_void_p
        pointer_to_env = c_void_p(self.libDSP.createEnv())
        return pointer_to_env

    def freeEnv(self, pointer_to_env: c_void_p) -> None:
        """freeEnv

        Frees a C++ DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a DSP environment

        Returns:

        """
        self.libDSP.freeEnv.argtypes = [c_void_p]
        self.libDSP.freeEnv(byref(pointer_to_env))

    def loadBlockProblem(self, pointer_to_env: c_void_p, id: int, ncols: int,
                         nrows: int, numels: int, start: list,
                         index: list, value: list, clbd: list, cubd: list,
                         coltype: list, obj: list, rlbd: list,
                         rubd: list) -> None:
        """loadBlockProblem

        Loads a block in block-structured problems. The loaded block is composed
         of a CSR form constraint matrix, an upper and lower bound on rows and
         columns, columns types and an objective vector.

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a DSP environment
            id (int) : number that identifies the block (0 is the master block)
            ncols (int) : total number of columns in the constraint matrix
            nrows (int) : number of lines in the loaded block constraint matrix
            numels (int) : number of nonzero values in the CSR matrix
            start (list <int>) : CSR format index pointer array (i.e., array
                                 storing the number of nonzero values in all
                                 previous rows of constraint matrix)
            index (list <int>) : CSR format index array (i.e., array storing
                                 column numbers of nonzero entries in
                                 constraint matrix)
            value (list <float>) : CSR format data array (i.e., array storing
                                   nonzero entries in constraint matrix)
            clbd (list <float>) : column lower bound (i.e., lower bound
                                  of corresponding variable)
            cubd (list <float>) : column upper bound (i.e., upper bound
                                  of corresponding variable)
            coltype (list <str>) : list of each columns' type
                                   ("I" for interger, "B" for binary and "C"
                                   for continuous, respectively)
            obj (list <float>) : objective vector (the objective vector must be
                                 the one of the master problem
                                 for the subproblems)
            rlbd (list <float>) : row lower bounds (i.e., lower bound of
                                  corresponding constraint)
            rubd (list <float>) : row upper bounds (i.e., upper bound of
                                  corresponding constraint)

        Returns:

        """

        self.libDSP.loadBlockProblem.argtypes = [c_void_p, c_int, c_int, c_int,
                                                 c_int,
                                                 POINTER(c_int),
                                                 POINTER(c_int),
                                                 POINTER(c_double),
                                                 POINTER(c_double),
                                                 POINTER(c_double),
                                                 POINTER(c_char),
                                                 POINTER(c_double),
                                                 POINTER(c_double),
                                                 POINTER(c_double)]
        c_id = c_int(id)
        c_ncols = c_int(ncols)
        c_nrows = c_int(nrows)
        c_numels = c_int(numels)

        c_start = self.__convert_list_to_c_int(start)
        c_index = self.__convert_list_to_c_int(index)
        c_value = self.__convert_list_to_c_double(value)
        c_clbd = self.__convert_list_to_c_double(clbd)
        c_cubd = self.__convert_list_to_c_double(cubd)
        c_coltype = self.__convert_column_type_to_c_char(coltype)
        c_obj = self.__convert_list_to_c_double(obj)
        c_rlbd = self.__convert_list_to_c_double(rlbd)
        c_rubd = self.__convert_list_to_c_double(rubd)

        self.libDSP.loadBlockProblem(pointer_to_env, c_id, c_ncols, c_nrows,
                                     c_numels, c_start, c_index, c_value,
                                     c_clbd, c_cubd, c_coltype, c_obj,
                                     c_rlbd, c_rubd)

    def printModel(self, pointer_to_env: c_void_p) -> None:
        """printModel

        Prints the model contained in the DSP environment pointer

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a DSP environment

        Returns:

        """
        self.libDSP.printModel.argtypes = [c_void_p]
        self.libDSP.printModel(pointer_to_env)

    def getCpuTime(self, pointer_to_env: c_void_p) -> float:
        """getCpuTime

        prints the resolution CPU time

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved DSP
                                        environment

        Returns:
            cpu_time (int) : CPU time spend in the solving
        """
        self.libDSP.getCpuTime.argtypes = [c_void_p]
        self.libDSP.getCpuTime.restype = c_double
        cpu_time = self.libDSP.getCpuTime(pointer_to_env)
        return cpu_time

    def getDualBound(self, pointer_to_env: c_void_p) -> float:
        """getDualBound

        Retrieves the dual bound from a solved DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved DSP
                                        environment

        Returns:
            dual_bound (float) : Dual bound of the solution
        """
        self.libDSP.getDualBound.argtypes = [c_void_p]
        self.libDSP.getDualBound.restype = c_double
        dual_bound = self.libDSP.getDualBound(pointer_to_env)
        return dual_bound

    def getDualSolution(self, pointer_to_env: c_void_p, ncols: int) -> list:
        """getDualSolution

        Retrieves the dual solution from a solved DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved DSP
                                        environment
            ncols (int) : number of columns in the optimization problem

        Returns:
            solution (list <float>) : Dual solution
        """
        self.libDSP.getDualSolution.argtypes = [c_void_p, c_int,
                                                POINTER(c_double)]
        c_ncols = c_int(ncols)
        c_solution = self.__convert_list_to_c_double([0] * ncols)
        self.libDSP.getDualSolution(pointer_to_env, c_ncols, c_solution)
        solution = self.__convert_to_list(c_solution)
        return solution

    def getPrimalBound(self, pointer_to_env: c_void_p) -> float:
        """getPrimalBound

        Retrieves the primal bound from a solved DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved DSP
                                        environment

        Returns:
            primal_bound (float) : Primal bound of the solution
        """
        self.libDSP.getPrimalBound.argtypes = [c_void_p]
        self.libDSP.getPrimalBound.restype = c_double
        primal_bound = self.libDSP.getPrimalBound(pointer_to_env)
        return primal_bound

    def getPrimalSolution(self, pointer_to_env: c_void_p, ncols: int) -> list:
        """getPrimalSolution

        Retrieves the primal solution from a solved DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved DSP
                                        environment
            ncols (int) : number of columns in the optimization problem

        Returns:
            solution (list <float>) : Primal solution
        """
        self.libDSP.getPrimalSolution.argtypes = [c_void_p, c_int,
                                                  POINTER(c_double)]
        c_ncols = c_int(ncols)
        c_solution = self.__convert_list_to_c_double([0] * ncols)
        self.libDSP.getPrimalSolution(pointer_to_env, c_ncols, c_solution)
        solution = self.__convert_to_list(c_solution)
        return solution

    def getNumIterations(self, pointer_to_env: c_void_p) -> int:
        """getNumIterations

        Retrieves the number of iterations needed to solve optimization problem

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved DSP
                                        environment

        Returns:
            nb_iteration (int) : Number of iterations taken
        """
        self.libDSP.getNumIterations.argtypes = [c_void_p]
        self.libDSP.getNumIterations.restype = c_int
        nb_iteration = self.libDSP.getNumIterations(pointer_to_env)
        return nb_iteration

    def getStatus(self, pointer_to_env: c_void_p) -> int:
        """getStatus

        Retrieves the status of a solved DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved
                                        DSP environment

        Returns:
            status (int) : solution status code
        """
        self.libDSP.getStatus.argtypes = [c_void_p]
        self.libDSP.getStatus.restype = c_int
        status = self.libDSP.getStatus(pointer_to_env)
        return status

    def getTotalNumCols(self, pointer_to_env: c_void_p) -> int:
        """getTotalNumCols

        Retrieves the total number of columns in a DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved
                                        DSP environment

        Returns:
            nb_col (int) : number of columns in the environment
        """
        self.libDSP.getTotalNumCols.argtypes = [c_void_p]
        self.libDSP.getTotalNumCols.restype = c_int
        nb_col = self.libDSP.getTotalNumCols(pointer_to_env)
        return nb_col

    def getTotalNumRows(self, pointer_to_env: c_void_p) -> int:
        """getTotalNumRows

        Retrieves the total number of rows in a DSP environment

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved
                                        DSP environment

        Returns:
            nb_rows (int) : number of rows in the environment
        """
        self.libDSP.getTotalNumRows.argtypes = [c_void_p]
        self.libDSP.getTotalNumRows.restype = c_int
        nb_rows = self.libDSP.getTotalNumRows(pointer_to_env)
        return nb_rows

    def getWallTime(self, pointer_to_env: c_void_p) -> int:
        """getWallTime

        Prints the solving wall time

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a solved
                                        DSP environment

        Returns:
            wall_time (int) : Wall time spend in the solving
        """
        self.libDSP.getWallTime.argtypes = [c_void_p]
        self.libDSP.getWallTime.restype = c_double
        wall_time = self.libDSP.getWallTime(pointer_to_env)
        return wall_time

    def updateBlocks(self, pointer_to_env: c_void_p) -> None:
        """updateBlocks

        Once all the blocks have been loaded, updateBlocks must be
        called to enable the solving

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a loaded
                                        DSP environment

        Returns:

        """
        self.libDSP.updateBlocks.argtypes = [c_void_p]
        self.libDSP.updateBlocks(pointer_to_env)

    def solveDe(self, pointer_to_env: c_void_p) -> None:
        """solveDe

        Solves an optimization problem using the Extensive Form method
        (i.e., aggregates all blocks and passes to off-the-shelf solver)

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a loaded DSP
            environment

        Returns:

        """
        self.libDSP.solveDe.argtypes = [c_void_p]
        self.libDSP.solveDe(pointer_to_env)

    def solveDw(self, pointer_to_env: c_void_p) -> None:
        """solveDw

        Solves an optimization problem using the Dantzig-Wolf algorithm

        Args:
            pointer_to_env (c_void_p) : ctypes pointer to a loaded
                                        DSP environment

        Returns:

        """
        self.libDSP.solveDw.argtypes = [c_void_p]
        self.libDSP.solveDw(pointer_to_env)

    @staticmethod
    def __convert_to_list(c_list: POINTER) -> list:
        """__convert_to_list

        Converts a list of ctypes elements to a Python list of elements

        Args:
            c_list : ctypes list of values

        Returns:
            list_values : Python list of values

        """
        list_values = []
        for c_value in c_list:
            list_values.append(c_value)
        return list_values

    @staticmethod
    def __convert_column_type_to_c_char(list_char: list) -> POINTER(c_char):
        """__convert_column_type_to_c_char

        Converts a list of strings (containing 'B', 'I' and 'C') to ctypes
        list of char equivalent

        Args:
            list_char : list of characters

        Returns:
            c_list_char : list of ctypes char

        """
        converted_char = []
        for char_i in list_char:
            if char_i == "I":
                converted_char.append(int(73))
            elif char_i == "B":
                converted_char.append(int(66))
            elif char_i == "C":
                converted_char.append(int(67))
            else:
                raise ValueError("Unknown character " + str(char_i))
        c_list_char = (c_char * len(converted_char))(*converted_char)
        return c_list_char

    @staticmethod
    def __convert_list_to_c_double(list_doubles: list) -> POINTER(c_double):
        """__convert_list_to_c_double

        Converts a list of floats to a ctypes list of c_double

        Args:
            list_doubles (list) : list of numbers

        Returns:
            c_list_doubles (POINTER(c_double) : list of ctypes c_double

        """
        c_list_doubles = (c_double * len(list_doubles))(*list_doubles)
        return c_list_doubles

    @staticmethod
    def __convert_list_to_c_int(list_ints: list) -> POINTER(c_int):
        """__convert_list_to_c_int

        Converts a list of floats to a ctypes list of c_int

        Args:
            list_ints : list of numbers

        Returns:
            c_list_ints : list of ctypes c_int

        """
        c_list_ints = (c_int * len(list_ints))(*list_ints)
        return c_list_ints



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .parent import Symbol
from gboml.compiler.utils import error_, list_to_string
import os


class Parameter(Symbol): 
    """
    Parameter object is composed of: 
    - a name 
    - a right handside expression or several expression
    """

    def __init__(self, name: str, expression, line=0):

        self.vector = None
        if expression is None:

            type_para = "table"
        elif type(expression) == str:

            type_para = "table"
            self.get_values_from_file(expression)
            expression = None
        else:

            type_para = "expression"
        Symbol.__init__(self, name, type_para, line)
        self.expression = expression
        self.value = None

    def __str__(self):
        
        string = "["+str(self.name)+' , '
        if self.expression is None:
            string += list_to_string(self.vector)
        else:
            string += str(self.expression)
        string += str(self.value)
        string += ']'
        
        return string

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value

    def get_values_from_file(self, expression):
        from .expression import Expression

        self.vector = []
        if type(expression) is str:

            if(os.path.isfile('./'+expression)) is False:

                error_("No such file as "+str(expression))
            f = open('./'+expression, "r", encoding="ISO-8859-1")
            for line in f:

                line = line.replace("\n", " ")
                line = line.replace(",", " ")
                line = line.replace(";", " ")
                line = line.split(" ")
                for nb in line:

                    if nb == "":

                        continue
                    try:

                        number = float(nb)
                        expr = Expression('literal', number)
                        self.vector.append(expr)
                    except ValueError:

                        error_("file "+expression
                               + " contains values that are not numbers "+nb)
            f.close()

    def get_expression(self):
        
        return self.expression

    def set_vector(self, v):
        
        self.vector = v

    def get_vector(self):
        
        return self.vector

    def get_number_of_values(self):

        if self.type == "expression":
            return 1
        else:
            return len(self.vector)



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""Solver API contains the API and option files from four different solvers.

All the APIs work simiarly, they take the matrix A, the vector b and vector c
as input of the problem
    min : c^T * X s.t. A * X <= b
and passes it to the corresponding solver.

  Args:
    A -> coo_matrix of constraints
    b -> np.ndarray of independent terms of each constraint
    c -> np.ndarray of objective vector
    objective_offset -> float of the objective offset
    name_tuples -> dictionary of <node_name variables> used to get the type

  Returns:
    solution -> np.ndarray of the flat solution
    objective -> float of the objective value
    status -> solution status
    solver_info -> dictionary of solver information

  Typical usage example:

   solution, objective, status, solver_info = api_solver(matrix_a, vector_b,
                                                        vector_c,
                                                        objective_offset,
                                                        name_tuples)
   print("the solution is "+str(solution))
   print("the objective found : "+str(objective))
"""

from .gurobi_solver import gurobi_solver
from .clp_solver import clp_solver
from .cplex_solver import cplex_solver
from .xpress_solver import xpress_solver
from .scipy_solver import scipy_solver
from .dsppy import DSPpy
from .dsp_solver import dsp_solver

__all__ = ["gurobi_solver", "clp_solver", "cplex_solver", "xpress_solver",
           "scipy_solver", "DSPpy", "dsp_solver"]



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

class Type:

    def __init__(self, type_id, line):
    
        self.type = type_id
        self.line = line

    def set_line(self, line):
    
        self.line = line

    def get_line(self):
    
        return self.line

    def get_type(self):
    
        return self.type

    def set_type(self, type_id):
    
        self.type = type_id


class Symbol(Type):
    
    def __init__(self, name, type_id, line):
    
        Type.__init__(self, type_id, line)
        if name is None:

            self.name = ""
        else:

            self.name = name

    def get_name(self):
    
        return self.name

    def set_name(self, name):
    
        self.name = name



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""DSP Solver file, contains the interface to DSP solver .

Takes the matrix A, the vector b and vector c as input of the problem
    min : c^T * X s.t. A * X <= b
and passes it to the dsp solver.

  Typical usage example:

   solution, objective, status, solver_info = dsp_solver(matrix_a, vector_b,
                                                        vector_c,
                                                        objective_offset,
                                                        name_tuples,
                                                        structure_indexes)
   print("the solution is "+str(solution))
   print("the objective found : "+str(objective))
"""

import numpy as np
from scipy.sparse import coo_matrix, csr_matrix

from gboml.compiler.utils import flat_nested_list_to_two_level
from .dsppy import DSPpy


def dsp_solver(matrix_a: coo_matrix, vector_b: np.ndarray, vector_c: np.ndarray,
               objective_offset: float, name_tuples: dict,
               structure_indexes,
               algorithm="de") -> tuple:
    """dsp_solver

        takes as input the matrix A, the vectors b and c. It returns the
        solution of the problem : min c^T * x s.t. A * x <= b found
        by the dsp solver

        Args:
            matrix_a -> coo_matrix of constraints
            vector_b -> np.ndarray of independent terms of each constraint
            vector_c -> np.ndarray of objective vector
            objective_offset -> float of the objective offset
            name_tuples -> dictionary of <node_name variables> used to get
                           the type
            structure_indexes -> constraint indexes of the different blocks
                                 (last one being the master block)
            algorithm -> solving algorithm to use, either "de" for the extensive
                         form and "dw" for Dantzig-Wolf

        Returns:
            solution -> np.ndarray of the flat solution
            objective -> float of the objective value
            status -> solution status
            solver_info -> dictionary of solver information

    """
    _, nb_cols = np.shape(matrix_a)
    all_constraints_matrix = csr_matrix(matrix_a)

    vector_c = vector_c[-1]

    start_master_constraint, end_master_constraint = structure_indexes[-1]
    master_csr = \
        all_constraints_matrix[start_master_constraint:
                               end_master_constraint + 1, 0:nb_cols]
    master_val, master_row, master_col = master_csr.data, master_csr.indptr,\
                                         master_csr.indices
    master_nb_lines = end_master_constraint + 1 - start_master_constraint
    minus_infinity_col = [-1000] * nb_cols
    plus_infinity_col = [1000] * nb_cols
    master_coltypes = ["C"]*nb_cols
    flat_name_tuples = flat_nested_list_to_two_level(name_tuples)

    for index, _, var_type, var_size in flat_name_tuples:

        if var_type == "integer":

            master_coltypes[index:index+var_size] = ["I"]*var_size
        if var_type == "binary":

            master_coltypes[index:index+var_size] = ["B"]*var_size

    row_lb = [-float("inf")] * master_nb_lines
    row_up = vector_b[start_master_constraint:end_master_constraint + 1]

    dsp = DSPpy()
    pointer_to_model = dsp.createEnv()
    dsp.loadBlockProblem(pointer_to_model, 0, nb_cols, master_nb_lines,
                         len(master_val), master_row, master_col, master_val,
                         minus_infinity_col, plus_infinity_col,
                         master_coltypes,
                         vector_c, row_lb, row_up)

    nb_blocks = len(structure_indexes)
    for b_number in range(nb_blocks-1):
        block_start_constr, block_end_constr = structure_indexes[b_number]
        block_nb_lines = block_end_constr - block_start_constr + 1
        block_csr = \
            all_constraints_matrix[block_start_constr:
                                   block_end_constr + 1, 0:nb_cols]
        block_val, block_row, block_col = block_csr.data, block_csr.indptr, \
                                          block_csr.indices
        row_up = vector_b[block_start_constr:block_end_constr + 1]
        row_lb = [-float("inf")] * block_nb_lines

        dsp.loadBlockProblem(pointer_to_model, b_number + 1, nb_cols,
                             block_nb_lines, len(block_val),
                             block_row, block_col, block_val,
                             minus_infinity_col,
                             plus_infinity_col, master_coltypes,
                             vector_c, row_lb, row_up)

    dsp.updateBlocks(pointer_to_model)
    # Retrieve solver information
    solver_info = {"name": "dsp", "algo": algorithm}
    solution = None
    objective = None

    # Solve the problem
    try:
        if algorithm == "de":
            dsp.solveDe(pointer_to_model)
        elif algorithm == "dw":
            dsp.solveDw(pointer_to_model)
        print("CPU Time : "+str(dsp.getCpuTime(pointer_to_model)))
        print("Wall Time : "+str(dsp.getWallTime(pointer_to_model)))

        status_code = dsp.getStatus(pointer_to_model)
        solver_info["status"] = status_code
        if status_code == 3000:

            status = "optimal"
            solution = dsp.getPrimalSolution(pointer_to_model, nb_cols)
            objective = dsp.getPrimalBound(pointer_to_model) + objective_offset
        else:

            status = "unknown"
    except RuntimeError as e:

        print(e)
        status = "error"
    dsp.freeEnv(pointer_to_model)
    return solution, objective, status, solver_info



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .parent import Symbol
from gboml.compiler.utils import error_
import copy


class Identifier(Symbol):
    """
    Identifier object is a structure composed of 
    - a type id
    - a name
    - an expression 
    - size ! NOT DONE
    - an unitialized index field for knowing its position 
      if it is a variable
    """

    def __init__(self, type_id: str, name_id: str, node_name: str = '',
                 expression=None, line: int = 0):

        assert type(name_id) == str, \
            "Internal error: expected string for identifier name"
        assert type_id == "basic" or type_id == "assign", \
            "Internal error: unknown type for identifier"
        Symbol.__init__(self, name_id, type_id, line)
        self.expression = expression
        self.node = None
        self.node_name = node_name
        self.index = 0  # GLOBAL INDEX inside the Ax <= b matrix
        self.size = None
        self.option = None

    def __str__(self):

        string = str(self.name)
        if self.node_name != "":
            string = str(self.node_name) + "," + string

        if self.expression is not None:
            string += str(self.expression)

        string = '['+string+']'

        return string

    def __copy__(self):

        expr = copy.copy(self.expression)
        identifier = Identifier(self.type, self.name, expression=expr)
        if self.node is not None:
            identifier.set_node(self.node)
        elif self.node_name != "":
            identifier.node_name = self.node_name

        return identifier

    def set_node(self, node):
        self.node = node
        self.node_name = node.get_name()

    def get_node(self):
        return self.node

    def get_node_name(self):
        return self.node_name

    def set_node_name(self, new_name):
        self.node_name = new_name

    def set_option(self, option):

        self.option = option

    def get_option(self):

        return self.option

    def set_size(self, dictionary):

        if self.expression is None:

            size = 1
        else:

            size = self.expression.evaluate_expression(dictionary)
        if size <= 0:
            error_('ERROR : wrong size for variable : '
                   + str(self) + " at line : " + str(self.get_line()))
        self.size = size

    def get_size(self):

        return self.size

    def name_compare(self, identifier_i) -> bool:

        # a == a[t] ! NO ANYMORE
        equal = False
        if type(identifier_i) == type(self):

            if self.name == identifier_i.name:
                equal = True
        elif type(identifier_i) == str:

            if self.name == identifier_i:
                equal = True

        return equal

    def set_expression(self, expr):

        self.expression = expr

    def get_expression(self):

        return self.expression

    def set_index(self, value):

        self.index = value

        return value + self.size

    def get_index(self):

        return self.index 



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .parent import Symbol
from .identifier import Identifier
from .error import RedefinitionError


class Variable(Symbol):
    """
    Variable object is composed of: 
    - a type 
    - an Identifier object
    """

    def __init__(self, identifier: Identifier,
                 v_type: str, v_option: str = "",
                 child_variable=None, line=0):
        assert type(v_type) == str, \
            "Internal error: expected string for variable type"
        assert v_type == "internal" or v_type == "external", \
            "Internal error: unknown variable type "+str(v_type)
        assert type(identifier) == Identifier, \
            "Internal error: identifier must be an Identifier object"
        Symbol.__init__(self, identifier, v_type, line)
        self.option = v_option
        self.dependencies = []
        self.lower_constraint = None
        self.upper_constraint = None
        self.initial_constraint = None
        self.dynamics = None
        self.assignment = None
        self.child_variable = child_variable

    def __str__(self):
        string = "[" + str(self.name) + ' , ' + str(self.type)
        string += ']'

        return string

    def get_child_assignment(self):
        return self.child_variable

    def get_lower_constraint(self):
        return self.lower_constraint

    def set_lower_constraint(self, expression):
        if self.lower_constraint is None:
            self.lower_constraint = expression
        else:
            raise RedefinitionError

    def get_upper_constraint(self):
        return self.upper_constraint

    def set_upper_constraint(self, expression):
        if self.upper_constraint is None:
            self.upper_constraint = expression
        else:
            raise RedefinitionError

    def get_dynamic(self):
        return self.dynamics

    def set_dynamic(self, dynamic):
        if self.dynamics is None:
            self.dynamics = dynamic
        else:
            raise RedefinitionError

    def get_assignment(self):
        return self.assignment

    def set_assignment(self, expression):
        if self.assignment is None:
            self.assignment = expression
        else:
            raise RedefinitionError

    def get_initial_constraint(self):
        return self.initial_constraint

    def set_initial_constraint(self, expression):
        if self.initial_constraint is None:
            self.initial_constraint = expression
        else:
            raise RedefinitionError

    def get_dependencies(self):
        return self.dependencies

    def set_dependencies(self, dependencies):
        self.dependencies = dependencies

    def get_option(self):
        return self.option

    def get_size(self):
        return self.name.get_size()

    def get_identifier(self):
        return self.get_name()

    def reset_type(self, vtype):
        assert vtype == "internal" or vtype == "external", \
            "Internal error: unknown variable type"
        self.type = vtype



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .compiler import parse_file, semantic, check_program_linearity, \
    matrix_generation_a_b, matrix_generation_c, \
    factorize_program, extend_factor, extend_factor_on_multiple_processes
from .compiler.classes import Parameter, Expression, Node, Hyperedge, Time, \
    Program
from .compiler.utils import error_, move_to_directory
from .solver_api import cplex_solver, gurobi_solver, clp_solver, dsp_solver, \
    xpress_solver
from .output import generate_json

from enum import Enum
import os
from copy import deepcopy
import numpy as np


class VariableType(Enum):
    EXTERNAL = "external"
    INTERNAL = "internal"


class GbomlGraph:
    """GbomlGraph makes it possible to define and solve a GBOML model.

    The GbomlGraph class enables the construction of GBOML models by
    importing nodes and hyperedges from a GBOML file. It also possesses
    a set of functions for updating the imported nodes and hyperedges
    (e.g., re-defining parameters or the type of variables). Nodes and
    hyperedges can be added to a GbomlGraph instance, from which the model
    can be solved and generated.

    Args:
        timehorizon (int) : length of optimization horizon considered

    :ivar list_nodes: nodes included in model
    :ivar list_hyperedges: hyperedges included in model
    :ivar timehorizon: optimization horizon object
    :ivar node_hyperedge_dict: dictionary of all nodes and hyperedges
    :ivar program: Program class of generated model (= None if model
                   not generated)
    :ivar matrix_a: constraint matrix A in sparse COO format (= None if model
                    not generated)
    :ivar matrix_b: upper bound on each row in constraint matrix (i.e.,
                    right-hand side coefficients, = None if model
                    not generated)
    :ivar vector_c: vector of objective coefficients (= None if model not
                    generated)
    :ivar indep_term_c: objective offset (i.e., constant term in the
                        objective, = None if model not generated)

    """
    def __init__(self, timehorizon=1):
        """
        bound method initializing a GbomlGraph instance

        Args:
           timehorizon (int) : length of optimization horizon considered

        """
        self.list_nodes = []
        self.list_hyperedges = []
        self.timehorizon: Time = Time("T", Expression('literal', timehorizon))
        self.node_hyperedge_dict = {}

        self.program = None
        self.matrix_a = None
        self.matrix_b = None
        self.vector_c = None
        self.indep_term_c = None

    def __add_node(self, to_add_node):
        """
        bound method adding a node to a GbomlGraph instance

        Args:
           to_add_node (Node) : node that should be added

        :raises: re-use of identifier error

        """
        node_name = to_add_node.get_name()
        if node_name in self.node_hyperedge_dict:
            error_("Re-use of same identifier twice for nodes or hyperedges : "
                   + str(node_name))
        self.node_hyperedge_dict[node_name] = to_add_node
        self.list_nodes.append(to_add_node)

    def __add_hyperedge(self, to_add_hyperedge):
        """
        bound method adding a hyperedge to a GbomlGraph instance

        Args:
           to_add_hyperedge (Hyperedge) : hyperedge that should be added

        :raises: re-use of identifier error

        """

        hyperedge_name = to_add_hyperedge.get_name()
        if hyperedge_name in self.node_hyperedge_dict:
            error_("Reuse of same identifier twice for nodes or hyperedges : "
                   + str(hyperedge_name))
        self.node_hyperedge_dict[hyperedge_name] = to_add_hyperedge
        self.list_hyperedges.append(to_add_hyperedge)

    def __get__(self, *node_or_hyperedge_name, wanted_type=None):
        """
        bound method returning a node or a hyperedge identified by its name
        and those of its ancestors

        Args:
           *node_or_hyperedge_name (list <str>) : list of ancestor node
                                                  names and node or
                                                  hyperedge name (used
                                                  for depth-first
                                                  traversal)
           wanted_type (Node or Hyperedge) : returned type (either Node
                                             or Hyperedge)

        :raises: not found error

        """

        retrieved_object = None
        current_layer = self.node_hyperedge_dict
        depth_search = len(node_or_hyperedge_name)
        for i, name in enumerate(node_or_hyperedge_name):
            if name in current_layer:
                retrieved_object = current_layer[node_or_hyperedge_name]
                if i < depth_search-1 and \
                        isinstance(retrieved_object, Hyperedge):
                    error_("Hyperedges do not possess subnodes or "
                           "subhyperedges")
                elif i < depth_search-1:
                    current_layer = retrieved_object.get_internal_dict()

            else:
                error_("Unknown node or hyperedge named "
                       + str(node_or_hyperedge_name))

        if wanted_type and not isinstance(retrieved_object, wanted_type):
            error_("Error wanted type " + str(wanted_type)
                   + " but that name references a " +
                   str(type(retrieved_object)))

        return retrieved_object

    def add_nodes_in_model(self, *nodes):
        """
        bound method adding nodes to a GbomlGraph instance

        Args:
            nodes (list <Nodes>) : list of node objects to be added

        """
        for node in nodes:
            self.__add_node(node)

    def add_hyperedges_in_model(self, *hyperedges):
        """
        bound method adding hyperedges to a GbomlGraph instance

        Args:
            hyperedges (list <Hyperedge>) : list of hyperedge objects to be
                                            added

        """
        for hyperedge in hyperedges:
            self.__add_hyperedge(hyperedge)

    def set_timehorizon(self, value):
        """
        bound method setting the time horizon to a specified value

        Args:
           value (int) : length of time horizon considered

        """
        self.timehorizon: Time = Time("T", Expression('literal', value))

    def get_timehorizon(self):
        """
        bound method returning the value of the time horizon

        Returns:
           value (int) : length of time horizon considered

        """

        return self.timehorizon.get_value()

    def build_model(self, nb_processes: int = 1):
        """
        bound method generating the matrices of the optimization model

        Args:
           nb_processes (int) : number of processes used for model generation

        """

        program = Program(self.list_nodes, timescale=self.timehorizon,
                          links=self.list_hyperedges)
        program, program_variables_dict, definitions = semantic(program)
        check_program_linearity(program, program_variables_dict, definitions)
        factorize_program(program, program_variables_dict, definitions)
        if nb_processes > 1:
            extend_factor_on_multiple_processes(program, definitions,
                                                nb_processes)
        else:
            extend_factor(program, definitions)

        matrix_a, vector_b = matrix_generation_a_b(program)
        vector_c, indep_terms_c = matrix_generation_c(program)
        program.free_factors_objectives()

        self.program = program
        self.matrix_a = matrix_a
        self.matrix_b = vector_b
        self.vector_c = vector_c
        self.indep_term_c = indep_terms_c

    def __solve(self, solver_function):
        """
        bound method solving the optimization model

        Args:
           solver_function (function) : function calling a solver

        Returns:
            solution (ndarray) : flattened solution
            objective (float) : objective value
            status (str) : solver exit status
            solver_info (dict) : dictionary storing solver information
            (*) constraints_information (dict) : dict of additional information
                                                 concerning constraints
            (*) variables_information (dict) : dict of additional information
                                               concerning variables

        Notes:
            (*) depending on the solver

        """
        vector_c = np.asarray(self.vector_c.sum(axis=0), dtype=float)
        objective_offset = float(self.indep_term_c.sum())
        return solver_function(self.matrix_a, self.matrix_b, vector_c,
                               objective_offset, self.program.get_tuple_name())

    def solve_gurobi(self, opt_file: str = None):
        """
        bound method solving the flattened optimization model with Gurobi

        Args:
            opt_file (str) : path to an optimization parameters file

        Returns:
            solution (ndarray) : flattened solution
            objective (float) : objective value
            status (str) : solver exit status
            solver_info (dict) : dictionary storing solver information
            constraints_information (dict) : dict of additional information concerning constraints
            variables_information (dict) : dict of additional information concerning variables

        """
        gurobi_solver_function = \
            lambda matrix_a, vector_b, vector_c, objective_offset, name_tuples:\
                gurobi_solver(matrix_a, vector_b, vector_c, objective_offset,
                              name_tuples, opt_file)
        return self.__solve(gurobi_solver_function)

    def solve_cplex(self, opt_file: str = None):
        """
        bound method solving the flattened optimization model with CPLEX

        Args:
            opt_file (str) : path to an optimization parameters file

        Returns:
            solution (ndarray) : flattened solution
            objective (float) : objective value
            status (str) : solver exit status
            solver_info (dict) : dictionary storing solver information
            constraints_information (dict) : dict of additional information concerning constraints
            variables_information (dict) : dict of additional information concerning variables

        """
        cplex_solver_function = \
            lambda matrix_a, vector_b, vector_c, objective_offset, name_tuples:\
                cplex_solver(matrix_a, vector_b, vector_c, objective_offset,
                              name_tuples, opt_file)
        return self.__solve(cplex_solver_function)

    def solve_xpress(self, opt_file: str = None):
        """
        bound method solving the flattened optimization model with Xpress

        Args:
            opt_file (str) : path to an optimization parameters file

        Returns:
            solution (ndarray) : flattened solution
            objective (float) : objective value
            status (str) : solver exit status
            solver_info (dict) : dictionary storing solver information
            constraints_information (dict) : dict of additional information concerning constraints
            variables_information (dict) : dict of additional information concerning variables
        """
        xpress_solver_function = \
            lambda matrix_a, vector_b, vector_c, objective_offset, name_tuples:\
                xpress_solver(matrix_a, vector_b, vector_c, objective_offset,
                              name_tuples, opt_file)
        return self.__solve(xpress_solver_function)

    def solve_clp(self):
        """
        bound method solving the flattened optimization problem with Clp/Cbc

        Returns:
            solution (ndarray) : flattened solution
            objective (flat) : float of the objective value
            status (str) : solver exit status
            solver_info (dict) : dictionary storing solver information

        """
        return self.__solve(clp_solver)

    def solve_dsp(self, algorithm="dw"):
        """
        bound method solving the optimization model with DSP

        Args:
            algorithm (str): algorithm selected ("dw" for Dantzig-Wolfe
                             and "de" for extensive form solve)

        Returns:
            solution (ndarray) : flattened solution
            objective (float ) : objective value
            status (str) : solver exit status
            solver_info (dict) : dictionary of solver information

        """

        dsp_solver_function = \
            lambda matrix_a, vector_b, vector_c, objective_offset, name_tuples:\
                dsp_solver(matrix_a, vector_b, vector_c, objective_offset,
                           name_tuples, self.program.
                           get_first_level_constraints_decomposition(),
                           algorithm=algorithm)

        return self.__solve(dsp_solver_function)

    def turn_solution_to_dictionary(self, solver_data, status, solution,
                                    objective, constraint_info=None,
                                    variables_info=None):
        """
        bound method converting the flat solution to a structured dictionary

        Args:
            solver_data (dict): dictionary of solver information
            status (str) : solver exit status
            solution (ndarray) : flattened solution
            objective (float ) : objective value
            constraint_info (dict) : dict of additional information
                                     concerning constraints
            variables_info (dict) : dict of additional information
                                    concerning variables

        Returns:
            gathered_data (dict): structured dictionary containing all the solution information
        """

        if constraint_info is None:
            constraint_info = dict()

        if variables_info is None:
            variables_info = dict()

        return generate_json(self.program, solver_data, status,
                             solution, objective, self.vector_c,
                             self.indep_term_c, constraint_info, variables_info)

    @staticmethod
    def import_all_nodes_and_edges(filename):
        """
        static method importing all nodes and hyperedges contained in a file

        Args:
           filename (str) : path to GBOML input file

        Returns:
            all_nodes (list) : list of nodes contained in file
            all_hyperedges (list) : list of hyperedges contained in file

        """
        old_dir, cut_filename = move_to_directory(filename)
        filename_graph = parse_file(cut_filename)
        all_nodes = filename_graph.get_nodes()
        to_return_nodes = []
        for node in all_nodes:
            to_return_nodes.append(deepcopy(node))

        all_hyperedges = filename_graph.get_links()
        to_return_edges = []
        for edge in all_hyperedges:
            to_return_edges.append(deepcopy(edge))
        os.chdir(old_dir)
        return to_return_nodes, to_return_edges

    @staticmethod
    def import_node(filename: str, *imported_node_identifier: str,
                    new_node_name: str = "", copy=True):
        """
        static method importing a node from a GBOML input file

        Args:
           filename (str) : path to GBOML input file
           imported_node_identifier (list <str>) : list of ancestor node
                                                   names and node name
                                                   (used for depth-first
                                                   traversal)
           new_node_name (str) : new identifier of node
                                 (for re-naming purposes, optional)
           copy (bool) : keyword argument defining whether a shallow or
                         deep copy of the imported node is created
                         (defaults to True, which produces a deepcopy)

        Returns:
            imported_node (Node) : imported node

        """
        old_dir, cut_filename = move_to_directory(filename)
        filename_graph = parse_file(cut_filename)
        imported_node = filename_graph.get(imported_node_identifier)

        if imported_node is None:
            error_("ERROR: In file " + str(filename) +
                   " there is no node named " + str(imported_node_identifier))

        if type(imported_node) != Node:
            error_("ERROR: A node named " +
                   str(imported_node_identifier) +
                   " is imported as type hyperedge ")

        if new_node_name == "":
            new_node_name = imported_node_identifier[-1]

        if copy:
            imported_node = deepcopy(imported_node)

        imported_node.rename(new_node_name)
        os.chdir(old_dir)
        return imported_node

    @staticmethod
    def import_hyperedge(filename: str, *imported_hyperedge_identifier: str,
                         new_hyperedge_name: str = "", copy=True):
        """
        static method importing a hyperedge from a GBOML input file

        Args:
           filename (str) : path to GBOML input file
           imported_hyperedge_identifier (list <str>) : list of ancestor node
                                                        names and hyperedge
                                                        name (used for depth-
                                                        first traversal)
           new_hyperedge_name (str) : new hyperedge identifier (for re-naming
                                      purposes, optional)
           copy (bool) : keyword argument defining whether a shallow or deep
                         copy of the imported node is created
                         (defaults to True, which produces a deepcopy)

        Returns:
            imported_hyperedge (Hyperedge) : imported hyperedge

        """
        old_dir, cut_filename = move_to_directory(filename)
        filename_graph = parse_file(filename)
        imported_hyperedge = filename_graph.get(imported_hyperedge_identifier)

        if imported_hyperedge is None:
            error_("ERROR: In file " + str(filename) +
                   " there is no node named " +
                   str(imported_hyperedge_identifier))

        if type(imported_hyperedge) == Hyperedge:
            error_("ERROR: A hyperedge named " +
                   str(imported_hyperedge_identifier)
                   + " is imported as type node ")

        if new_hyperedge_name == "":
            new_hyperedge_name = imported_hyperedge_identifier[-1]

        if copy:
            imported_hyperedge = deepcopy(imported_hyperedge)

        imported_hyperedge.rename(new_hyperedge_name)
        os.chdir(old_dir)
        return imported_hyperedge

    @staticmethod
    def rename(node_or_hyperedge, new_name):
        """
        static method re-naming a node or hyperedge

        Args:
           node_or_hyperedge (Node/Hyperedge) : node or hyperedge to
                                                be re-named
           new_name (str) : new name

        """
        node_or_hyperedge.rename(new_name)

    @staticmethod
    def get_object_in_node(in_node, *node_identifier: str, wanted_type=None):
        """
        static method returning a node or a hyperedge given an ancestor node

        Args:
           in_node (Node) : node to which the sub-node is expected to belong
           node_identifier (list <str>) : list of ancestor node names,
                                          with first name corresponding
                                          to first sub-node and last name
                                          corresponding to node to retrieve
           wanted_type (Class Type) : either Node or Hyperedge depending
                                      on the type of the object considered

        Returns:
            retrieved_object (Node/Hyperedge) : retrieved node or hyperedge

        """
        current_layer = in_node.get_internal_dict()
        depth_search = len(node_identifier)
        retrieved_object = None
        for i, name in enumerate(node_identifier):
            if name in current_layer:
                retrieved_object = current_layer[name]
                if i < depth_search - 1 and \
                        isinstance(retrieved_object, Hyperedge):
                    error_("Hyperedge objects do not possess subnodes "
                           "or subhyperedges")
                elif i < depth_search - 1:
                    current_layer = retrieved_object.get_internal_dict()

            else:
                error_("Unknown node or hyperedge named "
                       + str(node_identifier))

        if wanted_type and not isinstance(retrieved_object, wanted_type):
            error_("Error wanted type " + str(wanted_type)
                   + " but that name references a " +
                   str(type(retrieved_object)))

        return retrieved_object

    @staticmethod
    def add_sub_node(node_to_add, in_node):
        """
        static method adding a child node to a given node

        Args:
           node_to_add (Node) : sub-node to add
           in_node (Node) : node to which sub-node should be added

        """
        in_node.add_sub_node(node_to_add)
        in_node.update_internal_dict()

    @staticmethod
    def add_sub_hyperedge(hyperedge_to_add, in_node):
        """
        static method adding a sub-hyperedge to a given node

        Args:
           hyperedge_to_add (Hyperedge) : sub-hyperedge to add
           in_node (Node) : node to which sub-hyperedge should be added

        """
        in_node.add_link(hyperedge_to_add)
        in_node.update_internal_dict()

    @staticmethod
    def __redefine_parameter_from_value(node_or_hyperedge, parameter_name: str,
                                        value: float):
        """
        static method re-defining the value of a scalar parameter

        Args:
           node_or_hyperedge (Node/Hyperedge) : Node/Hyperedge where the
                                                parameter will be redefined
           parameter_name (str) : parameter name
           value (float) : new value of parameter

        """
        expr = Expression('literal', value)
        parameter = Parameter(parameter_name, expr)
        node_or_hyperedge.add_parameter_change(parameter)

    @staticmethod
    def __redefine_parameter_from_values(node_or_hyperedge, parameter_name: str,
                                         values: list):
        """
        static method re-defining parameter values from a list of values

        Args:
           node_or_hyperedge (Node/Hyperedge) : Node/Hyperedge to which the
                                                parameter that should be
                                                re-defined belongs
           parameter_name (str) : name of the parameter considered
           values (list<float>) : list of updated parameter values

        """

        expression_values = []
        for value in values:
            expr = Expression('literal', value)
            expression_values.append(expr)
        parameter = Parameter(parameter_name, None)
        parameter.set_vector(expression_values)
        node_or_hyperedge.add_parameter_change(parameter)

    @staticmethod
    def __redefine_parameter_from_file(node_or_hyperedge, parameter_name: str,
                                       filename):
        """
        static method re-defining parameter values from a CSV file

        Args:
           node_or_hyperedge (Node/Hyperedge) : Node/Hyperedge in which the
                                                parameter should be re-defined
           parameter_name (str) : parameter name
           filename (str) : name of CSV file

        """

        parameter = Parameter(parameter_name, filename)
        node_or_hyperedge.add_parameter_change(parameter)

    @staticmethod
    def redefine_parameters_from_list(node_or_hyperedge,
                                      list_parameters: list,
                                      list_values: list):
        """
        static method re-defining parameter values from a list

        Args:
           node_or_hyperedge (Node/Hyperedge) : Node/Hyperedge in which
                                                parameters should be re-defined
           list_parameters (list <str>) : list of parameter names
           list_values (list <float> | list <float> | <str>) : list of parameter
                                                               values

        """
        assert len(list_parameters) == len(list_values), \
            "Unmatching size between list or parameters and list of values"
        for i in range(len(list_parameters)):
            parameter_name = list_parameters[i]
            value = list_values[i]
            if isinstance(value, str):
                GbomlGraph.__redefine_parameter_from_file(node_or_hyperedge,
                                                          parameter_name,
                                                          value)
            elif isinstance(value, float) or isinstance(value, int):
                GbomlGraph.__redefine_parameter_from_value(node_or_hyperedge,
                                                           parameter_name,
                                                           value)
            elif isinstance(value, list):
                GbomlGraph.__redefine_parameter_from_values(node_or_hyperedge,
                                                            parameter_name,
                                                            value)
            else:
                error_("Unaccepted type value for parameter redefiniton "
                       + str(type(value)))

    @staticmethod
    def redefine_parameters_from_keywords(node_or_hyperedge, **kwargs):
        """
        static method re-defining parameter values from keyword arguments

        Args:
           node_or_hyperedge (Node/Hyperedge) : Node/Hyperedge in which
                                                parameters should be re-defined
           kwargs (tuple <str, value>): tuple of parameters name, value
        """
        for parameter_name, value in kwargs.items():
            if isinstance(value, str):
                GbomlGraph.__redefine_parameter_from_file(node_or_hyperedge,
                                                          parameter_name,
                                                          value)
            elif isinstance(value, float) or isinstance(value, int):
                GbomlGraph.__redefine_parameter_from_value(node_or_hyperedge,
                                                           parameter_name,
                                                           value)
            elif isinstance(value, list):
                GbomlGraph.redefine_parameter_from_values(node_or_hyperedge,
                                                          parameter_name,
                                                          value)
            else:
                error_("Unaccepted type value for parameter redefiniton "
                       + str(type(value)))

    @staticmethod
    def change_type_variable_in_node(node, variable_name: str, variable_type):
        """
        static method changing the type of a variable

        Args:
           node (Node) : node to which variable that should be modified belongs
           variable_name (str) : variable name
           variable_type (VariableType) : new variable type
                                          (either External or Internal)
        """
        variable_tuple = [variable_name, variable_type.value, 0]
        node.add_variable_change(variable_tuple)

    @staticmethod
    def remove_constraint(node_or_hyperedge, *to_delete_constraints_names):
        """
        static method removing constraints from a node/hyperedge

        Args:
           node_or_hyperedge (Node/Hyperedge) : Node/Hyperedge from which
                                                constraints should be removed
           to_delete_constraints_names (list <str>) : names of constraints
                                                      to remove
        """
        constraints = node_or_hyperedge.get_constraints()
        to_delete_constraints_names = list(to_delete_constraints_names)
        for constraint in constraints:
            constraint_name = constraint.get_name()
            if constraint_name in to_delete_constraints_names:
                node_or_hyperedge.remove_constraint(constraint)
                to_delete_constraints_names.remove(constraint_name)

        if to_delete_constraints_names:
            error_("Could not delete "+str(to_delete_constraints_names)
                   + " as they were not found in "
                   + str(node_or_hyperedge.get_name()))

    @staticmethod
    def remove_objective_in_node(node, *to_delete_objectives_names):
        """
        static method removing objectives from a node

        Args:
           node (Node/Hyperedge) : node from which objectives should be removed
           to_delete_objectives_names (list <str>) : names of objectives
                                                     to remove
        """
        objectives = node.get_objectives()
        to_delete_objectives_names = list(to_delete_objectives_names)
        for objective in objectives:
            objective_name = objective.get_name()
            if objective_name in to_delete_objectives_names:
                node.remove_objective(objective)
                to_delete_objectives_names.remove(objective_name)

        if to_delete_objectives_names:
            error_("Could not delete " + str(to_delete_objectives_names)
                   + " as they were not found in hyperedge "
                   + str(node.get_name()))

    @staticmethod
    def change_node_name_in_hyperedge(hyperedge, old_node_name, new_node_name):
        """
        static method changing the name of a node appearing in the constraints
        of a given hyperedge

        Args:
           hyperedge (Hyperedge) : Hyperedge in which node names should
                                   be changed
           old_node_name (str) : previous node name
           new_node_name (str) : new node name
        """
        change_tuple = [old_node_name, new_node_name, None]
        hyperedge.add_name_change(change_tuple)



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .parent import Type
from .expression import Expression


class Objective(Type):
    """
    Objective object is composed of: 
    - a type either min or max
    - an expression
    """

    def __init__(self, o_type, expression, time_interval=None,
                 condition=None, name=None, line=0):

        assert o_type == "min" or o_type == "max", \
            "Internal error: unknown objective type"
        assert type(expression) == Expression, \
            "Internal error: expected expression type expression in objective"
        Type.__init__(self, o_type, line)
        self.expression = expression
        self.time_interval = time_interval
        self.condition = condition
        self.name = name

    def __str__(self):

        string = "["+str(self.type)+','+str(self.expression)+']'
        
        return string

    def get_name(self):
        return self.name

    def get_expression(self):
        
        return self.expression

    def get_index_var(self):

        var_name = "t"
        if self.time_interval is not None:

            var_name = self.time_interval.get_index_name()
        
        return var_name

    def get_time_range(self, definitions):
        
        range_time = None
        if self.time_interval is not None:

            range_time = self.time_interval.get_range(definitions)
        
        return range_time

    def check_time(self, definitions):
        
        predicate = True
        if self.condition is not None:
            predicate = self.condition.check(definitions)
        
        return predicate

    def get_condition(self):

        return self.condition

    def get_time_interval(self):

        return self.time_interval



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""Scipy linprog Solver file, contains the interface to Linprog solver .

Takes the matrix A, the vector b and vector c as input of the problem
    min : c^T * X s.t. A * X <= b
and passes it to the linprog solver.

  Typical usage example:

   solution, objective, status, solver_info = scipy_solver(matrix_a, vector_b,
                                                           vector_c,
                                                           objective_offset,
                                                           name_tuples)
   print("the solution is "+str(solution))
   print("the objective found : "+str(objective))
"""

import numpy as np
from scipy.sparse import coo_matrix


def scipy_solver(matrix_a: coo_matrix, vector_b: np.ndarray,
                 vector_c: np.ndarray,
                 objective_offset: float, name_tuples: dict) -> tuple:
    """scipy_solver

        takes as input the matrix A, the vectors b and c. It returns the
        solution of the problem : min c^T * x s.t. A * x <= b found by the
        scipy's linprog solver

        Args:
            A -> coo_matrix of constraints
            b -> np.ndarray of independent terms of each constraint
            c -> np.ndarray of objective vector
            objective_offset -> float of the objective offset
            name_tuples -> dictionary of <node_name variables> used to get
                           the type

        Returns:
            solution -> np.ndarray of the flat solution
            objective -> float of the objective value
            status -> solution status
            solver_info -> dictionary of solver information

    """

    from scipy.optimize import linprog
    x0_bounds = (None, None)

    # Generate the model
    result = linprog(vector_c, A_ub=matrix_a.toarray(),
                     b_ub=vector_b,
                     bounds=x0_bounds,
                     options={"lstsq": True, "disp": True,
                              "cholesky": False, "sym_pos": False})

    # Retrieve solver info and solution
    solver_info = {"name": "linprog"}
    status_flag = result.success
    solver_info["status"] = status_flag
    solution = None
    objective = None
    if status_flag:

        status = "optimal"
        solution = result.x
        objective = result.fun+objective_offset
    else:

        status = "unknown"

    return solution, objective, status, solver_info
